from datetime import datetime
import numpy as np
from concert.tests import TestCase
from concert.quantities import q
from concert.devices.cameras.dummy import Camera, BufferedCamera
from concert.devices.cameras.pco import Timestamp, TimestampError


class TestDummyCamera(TestCase):

    def setUp(self):
        super(TestDummyCamera, self).setUp()
        self.background = np.ones((256, 256), dtype=np.uint16)
        self.camera = Camera(background=self.background)

    async def test_grab(self):
        frame = await self.camera.grab()
        self.assertIsNotNone(frame)

    def test_trigger_source(self):
        self.camera.trigger_source = self.camera.trigger_sources.EXTERNAL
        self.assertEqual(self.camera.trigger_source, 'EXTERNAL')

    def test_roi(self):
        self.roi_x0 = self.roi_y0 = self.roi_width = self.roi_height = 10 * q.dimensionless
        self.assertEqual(self.roi_x0, 10 * q.dimensionless)
        self.assertEqual(self.roi_y0, 10 * q.dimensionless)
        self.assertEqual(self.roi_width, 10 * q.dimensionless)
        self.assertEqual(self.roi_height, 10 * q.dimensionless)

    def test_has_sensor_dims(self):
        self.assertTrue(hasattr(self.camera, "sensor_pixel_width"))
        self.assertTrue(hasattr(self.camera, "sensor_pixel_height"))

    async def test_buffered_camera(self):
        camera = BufferedCamera()
        i = 0
        async for item in camera.readout_buffer():
            i += 1
        self.assertEqual(i, 3)

    async def test_context_manager(self):
        camera = Camera()

        async with camera.recording():
            self.assertEqual(await camera.get_state(), 'recording')
            await camera.grab()

        self.assertEqual(await camera.get_state(), 'standby')

    async def test_stream(self):
        async def check(producer):
            async for image in producer:
                check.ok = True
                await self.camera.stop_recording()
                break
        check.ok = False

        await check(self.camera.stream())
        self.assertTrue(check.ok)

    async def test_grab_convert(self):
        async def grab():
            return np.mgrid[:5, :5][1]

        self.camera._grab_real = grab
        self.camera.convert = np.fliplr
        image = await self.camera.grab()
        np.testing.assert_equal(image, (await grab())[:, ::-1])

    async def test_simulate(self):
        self.assertTrue(np.any(self.background - await self.camera.grab()))
        camera = Camera(background=self.background, simulate=False)
        np.testing.assert_equal(self.background, await camera.grab())


class TestPCOTimeStamp(TestCase):
    def test_valid(self):
        image = np.empty((1, 14), dtype=np.uint16)
        image[0] = np.array([0, 0, 0, 1, 32, 20, 8, 1, 20, 80, 69, 147, 40, 9], dtype=np.uint16)
        stamp = Timestamp(image)
        date_time = datetime(2014, 8, 1, 14, 50, 45, 932809)

        self.assertEqual(stamp.time, date_time)
        self.assertEqual(stamp.number, 1)

    def test_bad_input(self):
        image = np.empty((1,))
        with self.assertRaises(TypeError):
            Timestamp(image)

        image = np.empty((1,), dtype=np.uint16)
        with self.assertRaises(ValueError):
            Timestamp(image)

    def test_bad_data(self):
        image = np.empty((1, 14), dtype=np.uint16)
        image[0] = np.array([0, 0, 0, 1, 32, 20, 2**15, 1, 20, 80, 69, 147, 40, 9], dtype=np.uint16)
        with self.assertRaises(TimestampError):
            Timestamp(image)



"""
.. data:: MOTOR_VELOCITY_SAMPLING_TIME

    Time step for calculation of motor velocity by measuring two postion
    values. Longer values will create more acurate results but reading the
    velocity will take more time.

.. data:: PROGRESS_BAR

    Turn on progress bar by long-lasting operations if tqdm package is present
"""
from concert.quantities import q

# Prints the exception source by fake futures
PROGRESS_BAR = True

MOTOR_VELOCITY_SAMPLING_TIME = 0.1 * q.s

# Logging
AIODEBUG = 9
PERFDEBUG = 8



"""Dummy sample exchanger implementation."""

from concert.base import Selection
from concert.devices.samplechangers import base


class SampleChanger(base.SampleChanger):

    sample = Selection([None, 1, 2])

    def __init__(self):
        super(SampleChanger, self).__init__()
        self._sample = None

    async def _set_sample(self, sample):
        """Insert the sample in the holder."""
        self._sample = sample

    async def _get_sample(self):
        return self._sample



from concert.base import check, Selection
from concert.devices.lenschangers.base import LensChanger as BaseLensChanger


class LensChanger(BaseLensChanger):
    '''Lens changer class implementation'''
    objective = Selection(['objective_10x', 'objective_5x'], help='objective',
                          check=check(source='standby', target='standby'))

    def __init__(self):
        super(LensChanger, self).__init__()
        self._objective = 'objective_10x'

    async def _set_objective(self, objective):
        self._objective = objective

    async def _get_objective(self):
        return self._objective



"""Dummy"""
import asyncio
import logging
from concert.base import check, Parameter, Quantity, Selection, State, transition
from concert.coroutines.base import background
from concert.config import AIODEBUG
from concert.devices.base import Device
from concert.quantities import q


LOG = logging.getLogger(__name__)


async def set_evalue(instance, value):
    instance._value = value


async def get_evalue(instance):
    return instance._value


class DummyDevice(Device):

    """A dummy device."""

    position = Quantity(unit=q.mm)
    sleep_time = Quantity(unit=q.s)
    # Value with a check-decorated setter before it is bound to instance, so still a function
    value = Parameter()
    # value with check wrapping the bound method
    cvalue = Parameter(check=check(source='*', target='standby'))
    # Value set elsewhere
    evalue = Parameter(fset=set_evalue, fget=get_evalue, check=check(source='*', target='standby'))
    slow = Parameter()
    state = State(default='standby')

    def __init__(self, slow=None):
        super(DummyDevice, self).__init__()
        self._position = 1 * q.mm
        self._value = 0
        self._slow = slow
        self._sleep_time = 0.5 * q.s

    async def _get_sleep_time(self):
        return self._sleep_time

    async def _set_sleep_time(self, value):
        self._sleep_time = value

    async def _get_position(self):
        return self._position

    async def _set_position(self, value):
        self._position = value

    async def _get_slow(self):
        try:
            LOG.log(AIODEBUG, 'get slow start %s', self._slow)
            await asyncio.sleep((await self.get_sleep_time()).magnitude)
            LOG.log(AIODEBUG, 'get slow finish %s', self._slow)
            return self._slow
        except asyncio.CancelledError:
            LOG.log(AIODEBUG, 'get slow cancelled %s', self._slow)
            raise
        except KeyboardInterrupt:
            # do not scream
            LOG.debug("KeyboardInterrupt caught while getting")

    async def _set_slow(self, value):
        try:
            LOG.log(AIODEBUG, 'set slow start %s', value)
            await asyncio.sleep((await self.get_sleep_time()).magnitude)
            LOG.log(AIODEBUG, 'set slow finish %s', value)
            self._slow = value
        except asyncio.CancelledError:
            LOG.log(AIODEBUG, 'set slow cancelled %s', value)
            raise

    async def _get_value(self):
        """Get the real value."""
        return self._value

    async def _get_target_value(self):
        """Get the real value."""
        return self._value + 1

    @check(source='standby', target=['standby', 'hard-limit'])
    @transition(immediate='moving', target='standby')
    async def _set_value(self, value):
        """The real value setter."""
        self._value = value

    async def _get_cvalue(self):
        """The real value setter."""
        return self._value

    async def _set_cvalue(self, value):
        """The real value setter."""
        self._value = value

    @background
    async def do_nothing(self, value=None):
        """Do nothing. For testing task canellation."""
        await self._do_nothing(value=value)

    async def _do_nothing(self, value=None):
        LOG.log(AIODEBUG, f'Start doing nothing: {value}')
        try:
            await asyncio.sleep(1)
            LOG.log(AIODEBUG, f'Stop doing nothing: {value}')
            return value
        except asyncio.CancelledError:
            LOG.log(AIODEBUG, f'Doing nothing cancelled: {value}')
            raise

    async def _emergency_stop(self):
        LOG.debug('Emergency stop on a dummy device')
        await asyncio.sleep(0.5)
        self._state_value = 'aborted'


class SelectionDevice(Device):

    """A dummy device with a selection."""

    selection = Selection(list(range(3)))

    def __init__(self):
        super(SelectionDevice, self).__init__()
        self._selection = 0

    async def _get_selection(self):
        return self._selection

    async def _set_selection(self, selection):
        self._selection = selection



"""Image readers for convenient work with multi-page image sequences."""
import glob
import os
from concert.coroutines.base import run_in_executor


class FileSequenceReader:

    """Image sequence reader optimized for reading consecutive images. One multi-page image file is
    not closed after an image is read so that it does not have to be re-opened for reading the next
    image. The :func:`.close` function must be called explicitly in order to close the last opened
    image.
    """

    def __init__(self, file_prefix, ext=''):
        if os.path.isdir(file_prefix):
            file_prefix = os.path.join(file_prefix, '*' + ext)
        self._filenames = sorted(glob.glob(file_prefix))
        if not self._filenames:
            raise SequenceReaderError("No files matching `{}' found".format(file_prefix))
        self._lengths = {}
        self._file = None
        self._filename = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    async def read_range(self, start=0, stop=None, step=1):
        if stop is None:
            stop = self.num_images
        if stop > self.num_images:
            raise SequenceReaderError('Stop greater than number of images')

        try:
            for i in range(start, stop, step):
                yield await run_in_executor(self.read, i)
        finally:
            self.close()

    @property
    def num_images(self):
        num = 0
        for filename in self._filenames:
            num += self._get_num_images_in_file(filename)

        return num

    def read(self, index):
        if index < 0:
            # Enables negative indexing
            index += self.num_images
        file_index = 0
        while index >= 0:
            if file_index >= len(self._filenames):
                raise SequenceReaderError('image index greater than sequence length')
            index -= self._get_num_images_in_file(self._filenames[file_index])
            file_index += 1

        file_index -= 1
        index += self._lengths[self._filenames[file_index]]
        self._open(self._filenames[file_index])

        return self._read_real(index)

    def _open(self, filename):
        if self._filename != filename:
            if self._filename:
                self.close()
            self._file = self._open_real(filename)
            self._filename = filename

    def close(self):
        if self._filename:
            self._close_real()
            self._file = None
            self._filename = None

    def _get_num_images_in_file(self, filename):
        if filename not in self._lengths:
            self._open(filename)
            self._lengths[filename] = self._get_num_images_in_file_real()

        return self._lengths[filename]

    def _open_real(self, filename):
        """Returns an open file."""
        raise NotImplementedError

    def _close_real(self, filename):
        raise NotImplementedError

    def _get_num_images_in_file_real(self):
        raise NotImplementedError

    def _read_real(self, index):
        raise NotImplementedError


class TiffSequenceReader(FileSequenceReader):
    def __init__(self, file_prefix, ext='.tif'):
        super(TiffSequenceReader, self).__init__(file_prefix, ext=ext)

    def _open_real(self, filename):
        import tifffile
        return tifffile.TiffFile(filename)

    def _close_real(self):
        self._file.close()

    def _get_num_images_in_file_real(self):
        return len(self._file.pages)

    def _read_real(self, index):
        return self._file.pages[index].asarray()


class SequenceReaderError(Exception):
    pass



"""Storage Ring Device"""
from concert.quantities import q
from concert.base import Quantity, AccessorNotImplementedError
from concert.devices.base import Device


class StorageRing(Device):

    """Read-only access to storage ring information.

    .. py:attribute:: current

        Ring current

    .. py:attribute:: energy

        Ring energy

    .. py:attribute:: lifetime

        Ring lifetime in hours
    """
    current = Quantity(q.mA, help="Current")
    energy = Quantity(q.MeV, help="Energy")
    lifetime = Quantity(q.hour, help="Expected lifetime")

    def __init__(self):
        super(StorageRing, self).__init__()

    async def _get_current(self):
        raise AccessorNotImplementedError

    async def _get_energy(self):
        raise AccessorNotImplementedError

    async def _get_lifetime(self):
        raise AccessorNotImplementedError



"""Dummy scales."""
from concert.devices.scales import base
from concert.quantities import q


class Scales(base.Scales):

    """A dummy scale."""

    def __init__(self):
        super(Scales, self).__init__()
        self._weight = 147 * q.kg

    async def _get_weight(self):
        return self._weight


class TarableScales(base.TarableScales, Scales):

    """A tarable dummy scale."""

    async def _tare(self):
        pass



"""
Test experiments. Logging is disabled, so just check the directory and log
files creation.
"""
import asyncio
import numpy as np
import os.path as op
import tempfile
import shutil
from concert.quantities import q
from concert.coroutines.base import start
from concert.coroutines.sinks import Accumulate
from concert.experiments.base import Acquisition, Experiment, ExperimentError
from concert.experiments.imaging import (tomo_angular_step, tomo_max_speed,
                                         tomo_projections_number, frames)
from concert.experiments.addons import Addon, Consumer, ImageWriter, Accumulator
from concert.devices.cameras.dummy import Camera
from concert.tests import TestCase, suppressed_logging, assert_almost_equal
from concert.storage import DummyWalker, DirectoryWalker


class VisitChecker(object):

    """Use this to check that a callback was called."""

    def __init__(self):
        self.visited = False

    async def visit(self, *args, **kwargs):
        self.visited = True


class DummyAddon(Addon):
    def __init__(self):
        self.attached_num_times = 0
        super(DummyAddon, self).__init__([])

    def _attach(self):
        self.attached_num_times += 1

    def _detach(self):
        self.attached_num_times -= 1


class ExperimentSimple(Experiment):
    def __init__(self, walker):
        acq = Acquisition("test", self._run_test_acq)
        super(ExperimentSimple, self).__init__([acq], walker)

    async def _run_test_acq(self):
        for i in range(10):
            await asyncio.sleep(0.1)
            yield np.random.random((100, 100))


class ExperimentException(Experiment):
    def __init__(self, walker):
        acq = Acquisition("test", self._run_test_acq)
        super(ExperimentException, self).__init__([acq], walker)

    async def _run_test_acq(self):
        for i in range(2):
            yield i
        raise Exception("Run test acq")


@suppressed_logging
def test_tomo_angular_step():
    truth = np.arctan(2 / 100) * q.rad
    assert_almost_equal(truth, tomo_angular_step(100 * q.px))


@suppressed_logging
def test_projections_number():
    width = 100
    truth = int(np.ceil(np.pi / np.arctan(2 / width) * q.rad))
    assert truth == tomo_projections_number(width * q.px)


@suppressed_logging
def test_tomo_max_speed():
    width = 100
    frame_rate = 100 / q.s
    truth = np.arctan(2 / width) * q.rad * frame_rate
    assert_almost_equal(truth, tomo_max_speed(width * q.px, frame_rate))


class TestImagingFunctions(TestCase):

    async def test_frames(self):
        acc = Accumulate()
        await acc(frames(5, Camera()))
        assert len(acc.items) == 5


class TestAcquisition(TestCase):

    def setUp(self):
        self.acquired = False
        self.item = None
        self.acquisition = Acquisition('foo', self.produce, consumers=[self.consume],
                                       acquire=self.acquire)

    async def test_run(self):
        await self.acquisition()
        self.assertTrue(self.acquired)
        self.assertEqual(1, self.item)

    async def acquire(self):
        self.acquired = True

    async def produce(self):
        yield 1

    async def consume(self, producer):
        async for item in producer:
            self.item = item


class TestExperimentBase(TestCase):

    def setUp(self):
        super(TestExperimentBase, self).setUp()
        self.acquired = 0
        self.root = ''
        self.walker = DummyWalker(root=self.root)
        self.name_fmt = 'scan_{:>04}'
        self.visited = 0
        self.foo = Acquisition("foo", self.produce, consumers=[self.consume],
                               acquire=self.acquire)
        self.bar = Acquisition("bar", self.produce, acquire=self.acquire)
        self.acquisitions = [self.foo, self.bar]
        self.num_produce = 2
        self.item = None

    async def acquire(self):
        self.acquired += 1

    async def produce(self):
        self.visited += 1
        for i in range(self.num_produce):
            yield i

    async def consume(self, producer):
        async for item in producer:
            self.item = item


class TestExperiment(TestExperimentBase):

    def setUp(self):
        super(TestExperiment, self).setUp()
        self.experiment = Experiment(self.acquisitions, self.walker, name_fmt=self.name_fmt)
        self.visit_checker = VisitChecker()

    async def test_run(self):
        await self.experiment.run()
        self.assertEqual(self.visited, len(self.experiment.acquisitions))
        self.assertEqual(self.acquired, len(self.experiment.acquisitions))

        await self.experiment.run()
        self.assertEqual(self.visited, 2 * len(self.experiment.acquisitions))

        truth = set([op.join(self.root, self.name_fmt.format(i + 1)) for i in range(2)])
        self.assertEqual(truth, self.walker.paths)

        # Consumers must be called
        self.assertTrue(self.item is not None)

    def test_swap(self):
        self.experiment.swap(self.foo, self.bar)
        self.assertEqual(self.experiment.acquisitions[0], self.bar)
        self.assertEqual(self.experiment.acquisitions[1], self.foo)

    def test_get_by_name(self):
        self.assertEqual(self.foo, self.experiment.get_acquisition('foo'))
        self.assertRaises(ExperimentError, self.experiment.get_acquisition, 'non-existing')

    def test_acquisition_access(self):
        with self.assertRaises(AttributeError):
            self.experiment.acquisitions.remove(self.bar)

    def test_add(self):
        self.assertEqual(self.experiment.foo, self.foo)
        self.assertEqual(self.experiment.bar, self.bar)

    def test_remove(self):
        self.experiment.remove(self.bar)
        self.assertFalse(hasattr(self.experiment, 'bar'))
        self.assertNotIn(self.bar, self.experiment.acquisitions)

    async def test_prepare(self):
        self.experiment.prepare = self.visit_checker.visit
        await self.experiment.run()
        self.assertTrue(self.visit_checker.visited)

    async def test_finish(self):
        self.experiment.finish = self.visit_checker.visit
        await self.experiment.run()
        self.assertTrue(self.visit_checker.visited)

    async def test_consumer_addon(self):
        accumulate = Accumulate()
        Consumer([self.acquisitions[0]], accumulate)
        await self.experiment.run()
        self.assertEqual(accumulate.items, list(range(self.num_produce)))

    async def test_image_writing(self):
        ImageWriter(self.acquisitions, self.walker)
        await self.experiment.run()

        scan_name = self.name_fmt.format(1)
        # Check if the writing coroutine has been attached
        for i in range(self.num_produce):
            foo = op.join(self.root, scan_name, 'foo', self.walker.dsetname, str(i))
            bar = op.join(self.root, scan_name, 'bar', self.walker.dsetname, str(i))

            self.assertTrue(self.walker.exists(foo))
            self.assertTrue(self.walker.exists(bar))

    async def test_accumulation(self):
        acc = Accumulator(self.acquisitions)
        await self.experiment.run()

        for acq in self.acquisitions:
            self.assertEqual(acc.items[acq], list(range(self.num_produce)))

        # Test detach
        acc.detach()
        for acq in self.acquisitions:
            for consumer in acq.consumers:
                self.assertFalse(isinstance(consumer, Accumulate))
        self.assertEqual(acc.items, {})

    def test_attach_num_times(self):
        """An attached addon cannot be attached the second time."""
        addon = DummyAddon()
        # Does nothing because it's attached during construction
        addon.attach()
        self.assertEqual(1, addon.attached_num_times)

        # Detach
        addon.detach()
        self.assertEqual(0, addon.attached_num_times)

        # Second time, cannot be called
        addon.detach()
        self.assertEqual(0, addon.attached_num_times)


class TestExperimentStates(TestCase):
    def tearDown(self):
        shutil.rmtree(self.data_dir)

    def setUp(self):
        super(TestExperimentStates, self).setUp()
        self.data_dir = tempfile.mkdtemp()
        self.walker = DirectoryWalker(root=self.data_dir)

    async def test_experiment_state_normal(self):
        exp = ExperimentSimple(self.walker)
        self.assertEqual(await exp.get_state(), "standby")
        exp_handle = start(exp.run())
        await asyncio.sleep(0.2)
        self.assertEqual(await exp.get_state(), "running")
        await exp_handle
        self.assertEqual(await exp.get_state(), "standby")
        exp_handle = start(exp.run())
        await asyncio.sleep(0.2)
        self.assertEqual(await exp.get_state(), "running")
        exp_handle.cancel()
        await exp_handle
        self.assertEqual(await exp.get_state(), "standby")

    async def test_experiment_exception(self):
        exp = ExperimentException(self.walker)
        self.assertEqual(await exp.get_state(), "standby")
        with self.assertRaises(Exception):
            await exp.run()
        self.assertEqual(await exp.get_state(), "error")



# -*- coding: utf-8 -*-
"""Core module Parameters"""
import asyncio
import numpy as np
import logging
import functools
import inspect
import types
from concert.helpers import memoize
from concert.coroutines.base import background, run_in_loop, wait_until
from concert.quantities import q


LOG = logging.getLogger(__name__)


def identity(x):
    return x


def _setter_not_implemented(value, *args):
    raise AccessorNotImplementedError


def _getter_not_implemented(*args):
    raise AccessorNotImplementedError


def _getter_target_not_implemented(*args):
    raise AccessorNotImplementedError


async def _execute_func(func, instance, *args, **kwargs):
    """Execute *func* irrespective of whether it is a function or a method. *instance*
    is discarded if *func* is a function, otherwise it is used as a first real argument.
    """
    if isinstance(func, types.MethodType):
        result = await func(*args, **kwargs)
    else:
        result = await func(instance, *args, **kwargs)

    return result


def _find_object_by_name(instance):
    """Find variable name by *instance*. This is supposed to be used only in the
    :meth:`.Parameter.__set__`.
    """
    def _find_in_dict(dictionary):
        for (obj_name, obj) in list(dictionary.items()):
            if obj is instance:
                return obj_name

    instance_name = None

    frames = inspect.stack()
    try:
        # Skip us and Parameter.__set__
        for i in range(2, len(frames)):
            # First look in the globals
            instance_name = _find_in_dict(frames[i][0].f_globals)
            if not instance_name:
                # If not found, look in the locals (e.g. devices instantiated in funcions)
                instance_name = _find_in_dict(frames[i][0].f_locals)
            # We can't know at which index of the stack is the correct name, e.g. if a device sets a
            # parameter in a constructor we need to bump index by one. Thus, use blacklist names
            # which won't be picked up because they are most probably used by concert or it's
            # extensions internally.
            if instance_name and instance_name not in ['instance', 'self']:
                break
    finally:
        # Cleanup as python docs suggest
        del frames

    return instance_name


class FSMError(Exception):
    """All errors connected with the finite state machine"""


class TransitionNotAllowed(FSMError):
    pass


class StateError(Exception):

    """Raised in state check functions of devices."""

    def __init__(self, error_state, msg=None):
        self.state = error_state
        self.msg = msg if msg else "Got into `{}'".format(error_state)

    def __str__(self):
        return self.msg


class UnitError(ValueError):

    """Raised when an operation is passed value with an incompatible unit."""
    pass


class LimitError(ValueError):

    """Raised when an operation is passed a value that exceeds a limit."""
    pass


class SoftLimitError(LimitError):

    """Raised when a soft limit is hit on the device."""
    pass


class HardLimitError(StateError):

    """Raised when device goes into hardlimit error state."""

    def __init__(self, error_state='hard-limit', msg=None):
        super(HardLimitError, self).__init__(error_state, msg)


class ParameterError(Exception):

    """Raised when a parameter is accessed that does not exists."""

    def __init__(self, parameter):
        msg = "`{0}' is not a parameter".format(parameter)
        super(ParameterError, self).__init__(msg)


class AccessorNotImplementedError(NotImplementedError):

    """Raised when a setter or getter is not implemented."""


class ReadAccessError(Exception):

    """Raised when user tries to read a parameter that cannot be read."""

    def __init__(self, parameter):
        LOG.warn("Invalid read access on {0}".format(parameter))
        msg = "parameter `{0}' cannot be read".format(parameter)
        super(ReadAccessError, self).__init__(msg)


class TargetAccessError(Exception):

    """Raised when user tries to read parameter's target value that cannot be read."""

    def __init__(self, parameter):
        LOG.warn("Invalid read access on {0}".format(parameter))
        msg = "parameter's `{0}' target value cannot be read".format(parameter)
        super(TargetAccessError, self).__init__(msg)


class WriteAccessError(Exception):

    """Raised when user tries to read a parameter that cannot be read."""

    def __init__(self, parameter):
        LOG.warn("Invalid write access on {0}".format(parameter))
        msg = "parameter `{0}' cannot be written".format(parameter)
        super(WriteAccessError, self).__init__(msg)


class SelectionError(Exception):

    """Raised when user tries to set selection parameter to unknown value."""

    def __init__(self, parameter, value, values):
        LOG.warn("Invalid write access on {0}".format(parameter))
        msg = "Value `{}' not in `{}'".format(value, values)
        super(SelectionError, self).__init__(msg)


class LockError(Exception):

    """Raised when parameter is locked."""

    pass


def transition(immediate=None, target=None):
    """Change software state of a device to *immediate*. After the function execution finishes
    change the state to *target*. On :py:class:`.asyncio.CancelledError`, state is set to *target*
    and cleanup logic must take place in the callable to be wrapped.
    """
    def wrapped(func):
        @functools.wraps(func)
        async def call_func(instance, *args, **kwargs):
            if 'state' not in instance:
                raise FSMError('Changing state requires state parameter')

            # Store the original in case target is None
            target_state = target if target else await instance['state'].get()

            if immediate:
                setattr(instance, '_state_value', immediate)

            try:
                result = await _execute_func(func, instance, *args, **kwargs)
                setattr(instance, '_state_value', target_state)
            except StateError as error:
                setattr(instance, '_state_value', error.state)
                raise error
            except asyncio.CancelledError:
                # If *func* is cancelled, *func* must make sure that all cleanup except state update
                # happens. If even the state should have some special value, then do not use
                # *transition* at all and implement everything in *func*.
                setattr(instance, '_state_value', target_state)
                raise

            return result

        return call_func

    return wrapped


def check(source='*', target='*'):
    """
    Decorates a method for checking the device state.

    *source* denotes the source state that must be present at the time of
    invoking the decorated method. *target* is the state that the state object
    will be after successful completion of the method or a list of possible
    target states.
    """
    async def check_now(instance, allowed_states, state_str):
        state = await instance['state'].get()
        if state not in allowed_states and '*' not in allowed_states:
            raise TransitionNotAllowed(f"{state_str} state `{state}' not in `{allowed_states}'")

    def wrapped(func):
        sources = [source] if isinstance(source, str) else source
        targets = [target] if isinstance(target, str) else target

        @functools.wraps(func)
        async def call_func(instance, *args, **kwargs):
            if 'state' not in instance:
                raise FSMError('Transitioning requires state parameter')

            await check_now(instance, sources, 'Current')
            result = await _execute_func(func, instance, *args, **kwargs)
            await check_now(instance, targets, 'Final')

            return result

        return call_func

    return wrapped


class Parameter(object):

    """A parameter with getter and setter.

    Parameters are similar to normal Python properties and can additionally
    trigger state checks. If *fget* or *fset* is not given, you must
    implement the accessor functions named `_set_name` and `_get_name`::

        from concert.base import Parameter, State, check

        class SomeClass(object):

            state = State(default='standby')
            param = Parameter(check=check(source='standby', target=['standby', 'moving']))

            def _set_param(self, value):
                pass

            def _get_param(self):
                pass

    When a :class:`.Parameter` is attached to a class, you can modify it by
    accessing its associated :class:`.ParameterValue` with a dictionary
    access::

        obj = SomeClass()
        print(obj['param'])
    """

    def __init__(self, fget=None, fset=None, fget_target=None, data=None, check=None, help=None):
        """
        *fget* is a callable that is called when reading the parameter. *fset*
        is called when the parameter is written to. *fget_target* is a getter for the target value.
        *fget*, *fset*, *fget_target* must be member functions of the corresponding Parameterizable
        object.

        *data* is passed to the state check function.

        *check* is a :func:`.check` that changes states when a value
        is written to the parameter.

        *help* is a string describing the parameter in more detail.
        """
        self.name = None
        self.fget = fget
        self.fset = fset
        self.fget_target = fget_target
        self.data_args = (data,) if data is not None else ()
        self.check = check
        self.decorated = None
        self.help = help

    @memoize
    def setter_name(self):
        if self.fset:
            return self.fset.__name__

        return '_set_' + self.name

    @memoize
    def getter_name(self):
        if self.fget:
            return self.fget.__name__

        return '_get_' + self.name

    @memoize
    def getter_name_target(self):
        if self.fget_target:
            return self.fget_target.__name__

        return '_get_target_' + self.name

    @memoize
    def get_getter(self, instance):
        if self.fget:
            getter = functools.partial(self.fget, instance)
        else:
            getter = getattr(instance, self.getter_name())

        return getter

    @memoize
    def get_target_getter(self, instance):
        if self.fget_target:
            target_getter = functools.partial(self.fget_target, instance)
        else:
            target_getter = getattr(instance, self.getter_name_target())

        return target_getter

    @memoize
    def get_setter(self, instance):
        if self.fset:
            # If check is supplied to the Parameter do not apply partial because it will be applied
            # below
            setter = self.fset if self.check else functools.partial(self.fset, instance)
        else:
            setter = getattr(instance, self.setter_name())

        if self.check:
            setter = self.check(setter)
            # Check wraps a method into a function which expects the first argument to be instance
            # again, thus wrap it by functools.partial
            setter = functools.partial(setter, instance)

        return setter

    def __repr__(self):
        return str(self.help)

    def __get__(self, instance, owner):
        return run_in_loop(instance[self.name].get())

    def __set__(self, instance, value):
        run_in_loop(instance[self.name].set(value))


class State(Parameter):

    """
    Finite state machine.

    Use this on a class, to keep some sort of known state. In order to enforce
    restrictions, you would decorate methods on the class with
    :func:`.check`::

        class SomeObject(object):

            state = State(default='standby')

            @check(source='*', target='moving')
            def move(self):
                pass

    In case your device doesn't provide information on its state you can use
    the :func:`.transition` to store the state in an instance of your device::

        @transition(immediate='moving', target='standby')
        def _set_some_param(self, param_value):
            # when the method starts device state is set to *immediate*
            # long operation goes here
            pass
            # the state is set to *target* in the end

    Accessing the state variable will return the current state value, i.e.::

        obj = SomeObject()
        assert obj.state == 'standby'

    The state cannot be set explicitly by::

        obj.state = 'some_state'

    but the object needs to provide methods which transition out of
    states, the same holds for transitioning out of error states.
    If the :meth:`_get_state` method is implemented in the device
    it is always used to get the state, otherwise the state is stored
    in software.
    """

    def __init__(self, default=None, fget=None, fset=None, data=None, check=None, help=None):
        super(State, self).__init__(fget=fget, help=help)
        self.default = default

    def __set__(self, instance, value):
        raise AttributeError('State cannot be set')

    def _value(self, instance):
        if not hasattr(instance, '_state_value'):
            setattr(instance, '_state_value', self.default)

        return getattr(instance, '_state_value')


class Selection(Parameter):

    """A :class:`.Parameter` that can take a value out of pre-defined list."""

    def __init__(self, iterable, fget=None, fset=None, check=None, help=None):
        """
        *fget*, *fset*, *check* and *help* are identical to the :class:`.Parameter` constructor
        arguments.

        *iterable* is the list of things, that a selection can be.
        """
        super(Selection, self).__init__(fget=fget, fset=fset, check=check, help=help)
        self.iterable = iterable


class Quantity(Parameter):

    """A :class:`.Parameter` associated with a unit."""

    def __init__(self, unit, fget=None, fset=None, fget_target=None, lower=None, upper=None,
                 data=None, check=None, external_lower_getter=None, external_upper_getter=None,
                 help=None):
        """
        *fget*, *fset*, *data*, *check* and *help* are identical to the
        :class:`.Parameter` constructor arguments.

        *unit* is a Pint quantity. *lower* and *upper* denote soft limits
        between the :class:`.Quantity` values can lie.
        """
        super(Quantity, self).__init__(fget=fget, fset=fset, fget_target=fget_target, data=data,
                                       check=check, help=help)
        self.unit = unit

        self.upper = upper
        self.lower = lower
        self.external_lower_getter = external_lower_getter
        self.external_upper_getter = external_upper_getter

    def convert(self, value):
        return value.to(self.unit)


def quantity(unit=None, lower=None, upper=None, data=None, check=None, help=None):
    """
    Decorator for read-only quantity functions.

    Device authors can add additional read-only quantities to a specific device
    by applying this decorator to a function::

        class SomeDevice(Device):
            @quantity(unit=1*q.mm, lower=0*q.mm, upper=2*q.mm)
            def position(self):
                pass

    The arguments correspond to those of :class:`.Quantity`.
    """

    def wrapper(func):
        # Get help info from doc string if no explicit help was passed.
        doc = help if help else inspect.getdoc(func)

        return Quantity(fget=func, unit=unit, lower=lower, upper=upper,
                        data=data, check=check, help=doc)

    return wrapper


class ParameterValue(object):

    """Value object of a :class:`.Parameter`."""

    def __init__(self, instance, parameter):
        self._lock = asyncio.Lock()
        self._locked = False
        self._instance = instance
        self._parameter = parameter
        self._saved = []

    async def __aenter__(self):
        await self._lock.acquire()

        return self

    async def __aexit__(self, exc_type, exc, tb):
        self._lock.release()

    def __lt__(self, other):
        return self._parameter.name < other._parameter.name

    def __repr__(self):
        return run_in_loop(self.info_table).get_string()

    @property
    async def info_table(self):
        from concert.session.utils import get_default_table
        locked = "yes" if self.locked else "no"
        table = get_default_table(["attribute", "value"])
        table.header = False
        table.border = False
        table.add_row(["info", self._parameter.help])
        table.add_row(["locked", locked])
        table.add_row(["target_readable", self.target_readable])
        if self.target_readable:
            table.add_row(["target_value", await self.get_target()])
        return table

    @property
    def name(self):
        return self._parameter.name

    @property
    def data(self):
        return self._parameter.data

    @property
    def writable(self):
        """Return True if the parameter is writable."""
        return (getattr(self._instance, '_set_' + self.name) is not _setter_not_implemented
                or self._parameter.fset is not None)

    @property
    def target_readable(self):
        return (getattr(self._instance, '_get_target_' + self.name)
                is not _getter_target_not_implemented or self._parameter.fget_target is not None)

    @property
    def target(self):
        return run_in_loop(self.get_target())

    @background
    async def get(self, wait_on=None):
        """
        Get coroutine obtaining the concrete *value* of this object.

        If *wait_on* is not None, it must be an awaitable on which this method waits.
        """
        if wait_on:
            await wait_on

        try:
            getter = self._parameter.get_getter(self._instance)
            return await getter(*self._parameter.data_args)
        except asyncio.CancelledError:
            LOG.debug('getter cancelled %s', self._parameter.name)
        except AccessorNotImplementedError:
            raise ReadAccessError(self.name)

    @background
    async def get_target(self, wait_on=None):
        """
        Get coroutine obtaining target value of this object.

        If *wait_on* is not None, it must be an awaitable on which this method waits.
        """
        if wait_on:
            await wait_on

        try:
            target_getter = self._parameter.get_target_getter(self._instance)
            return await target_getter()
        except AccessorNotImplementedError:
            raise TargetAccessError(self.name)

    @background
    async def set(self, value, wait_on=None):
        """
        Set concrete *value* on the object.

        If *wait_on* is not None, it must be an awaitable on which this method waits.
        """
        if wait_on:
            await wait_on

        if self.locked:
            raise LockError("Parameter `{}' is locked for writing".format(self._parameter.name))

        try:
            setter = self._parameter.get_setter(self._instance)
            await setter(value, *self._parameter.data_args)
        except AccessorNotImplementedError:
            raise WriteAccessError(self.name)

        name = self._instance.__class__.__name__
        if hasattr(self._instance, 'name_for_log'):
            instance_name = getattr(self._instance, 'name_for_log')
        else:
            instance_name = _find_object_by_name(self._instance)
            if instance_name:
                setattr(self._instance, 'name_for_log', instance_name)
        if instance_name:
            msg = "set {}::{}.{}='{}'"
            LOG.info(msg.format(name, instance_name, self._parameter.name, value))
        else:
            msg = "set {}::{}='{}'"
            LOG.info(msg.format(name, self._parameter.name, value))

    @background
    async def stash(self):
        """Save the current value internally on a growing stack.

        If the parameter is writable the current value is saved on a stack and
        to be later retrieved with :meth:`.ParameterValue.restore`.
        """
        if not self.writable:
            raise ParameterError("Parameter `{}' is not writable".format(self.name))

        if self.target_readable:
            self._saved.append(await self.get_target())
        else:
            self._saved.append(await self.get())

    @background
    async def restore(self):
        """Restore the last value saved with :meth:`.ParameterValue.stash`.

        If the parameter can only be read or no value has been saved, this
        operation does nothing.
        """
        if not self.writable:
            raise ParameterError("Parameter `{}' is not writable".format(self.name))

        if self._saved:
            val = self._saved.pop()
            await self.set(val)

    @property
    def locked(self):
        """Return True if the parameter is locked for writing."""
        return self._locked

    def lock(self, permanent=False):
        """Lock parameter for writing. If *permament* is True the parameter
        cannot be unlocked anymore.
        """
        def unlock_not_allowed():
            raise LockError("Parameter `{}' cannot be unlocked".format(self))

        self._locked = True
        if permanent:
            self.unlock = unlock_not_allowed

    def unlock(self):
        """Unlock parameter for writing."""
        self._locked = False

    @background
    async def wait(self, value, sleep_time=1e-1 * q.s, timeout=None):
        """Wait until the parameter value is *value*. *sleep_time* is the time to sleep
        between consecutive checks. *timeout* specifies the maximum waiting time.
        """
        async def condition():
            return await self.get() == value

        await wait_until(condition, sleep_time=sleep_time, timeout=timeout)


class StateValue(ParameterValue):

    """Special :class:`.ParameterValue` implementing state parameter."""

    @background
    async def get(self, wait_on=None):
        if wait_on:
            await wait_on

        try:
            getter = self._parameter.get_getter(self._instance)
            return await getter(*self._parameter.data_args)
        except AccessorNotImplementedError:
            if not self._parameter.fget:
                if self._parameter.default is None:
                    raise FSMError('Software state must have a default value')
                if not hasattr(self._instance, '_state_value'):
                    self._instance._state_value = self._parameter.default
                return self._instance._state_value
            else:
                raise ReadAccessError(self.name)


class QuantityValue(ParameterValue):

    def __init__(self, instance, quantity):
        super(QuantityValue, self).__init__(instance, quantity)
        self._lower = quantity.lower
        self._upper = quantity.upper
        self._external_upper_getter = quantity.external_upper_getter
        self._external_lower_getter = quantity.external_lower_getter
        self._limits_locked = False

    def lock_limits(self, permanent=False):
        """Lock limits, if *permanent* is True the limits cannot be unlocked anymore."""
        def unlock_not_allowed():
            raise LockError('Limits are locked permanently')

        self._limits_locked = True
        if permanent:
            self.unlock_limits = unlock_not_allowed

    def unlock_limits(self):
        """Unlock limits."""
        self._limits_locked = False

    @property
    def lower(self):
        if self.lower_user is None and self.lower_external is None:
            return None
        elif self.lower_user is None and self.lower_external is not None:
            return self.lower_external
        elif self.lower_user is not None and self.lower_external is None:
            return self.lower_user
        else:
            return np.max((self.lower_user.to(self._parameter.unit).magnitude,
                           self.lower_external.to(self._parameter.unit).magnitude),
                          axis=0) * self._parameter.unit

    @lower.setter
    def lower(self, value):
        if value is None:
            self._lower = None
            return
        self._check_limit(value)
        if self._upper is not None and value >= self._upper:
            raise ValueError('Lower limit must be lower than upper')
        self._lower = value

    @property
    def upper(self):
        if self.upper_user is None and self.upper_external is None:
            return None
        elif self.upper_user is None and self.upper_external is not None:
            return self.upper_external
        elif self.upper_user is not None and self.upper_external is None:
            return self.upper_user
        else:
            return np.min((self.upper_user.to(self._parameter.unit).magnitude,
                           self.upper_external.to(self._parameter.unit).magnitude),
                          axis=0) * self._parameter.unit

    @upper.setter
    def upper(self, value):
        if value is None:
            self._upper = None
            return
        self._check_limit(value)
        if self._lower is not None and value <= self._lower:
            raise ValueError('Upper limit must be greater than lower')
        self._upper = value

    @property
    def upper_user(self):
        return self._upper

    @property
    def lower_user(self):
        return self._lower

    @property
    def lower_external(self):
        if self._external_lower_getter is None:
            return None
        else:
            return self._external_lower_getter()

    @property
    def upper_external(self):
        if self._external_upper_getter is None:
            return None
        else:
            return self._external_upper_getter()

    @property
    async def info_table(self):
        table = await super(QuantityValue, self).info_table
        table.add_row(["lower", self.lower])
        table.add_row(["upper", self.upper])

        if self._external_lower_getter is not None:
            table.add_row(["lower_user", self.lower_user])
            table.add_row(["lower_external", self.lower_external])
        if self._external_upper_getter is not None:
            table.add_row(["upper_user", self.upper_user])
            table.add_row(["upper_external", self.upper_external])
        return table

    @property
    def unit(self):
        return self._parameter.unit

    @background
    async def get(self, wait_on=None):
        value = await super(QuantityValue, self).get(wait_on=wait_on)

        if value is not None:
            return self._parameter.convert(value)

    @background
    async def set(self, value, wait_on=None):
        """
        Set concrete *value* on the object.

        If *wait_on* is not None, it must be an awaitable on which this method waits.
        """
        if wait_on:
            await wait_on

        if not self.unit.is_compatible_with(value):
            msg = "{} of {} can only receive values of unit {} but got {}"
            raise UnitError(
                msg.format(self._parameter.name, type(self._instance), self.unit, value))

        def leq(a, b):
            """check a <= b"""
            if not hasattr(a, 'shape') or len(a.shape) == 0:
                return a <= b
            else:
                # Vector data type
                valid_a = np.invert(np.isnan(a.magnitude))
                valid_b = np.invert(np.isnan(b.magnitude))
                valid = valid_a & valid_b
                test_a = a[valid]
                test_b = b[valid]
                return np.all(test_a.to_base_units().magnitude <= test_b.to_base_units().magnitude)

        if self.lower is not None:
            if not leq(self.lower, value):
                msg = "{} is out of range [{}, {}]"
                raise SoftLimitError(msg.format(value, self.lower, self.upper))
        if self.upper is not None:
            if not leq(value, self.upper):
                msg = "{} is out of range [{}, {}]"
                raise SoftLimitError(msg.format(value, self.lower, self.upper))

        converted = self._parameter.convert(value)
        await super(QuantityValue, self).set(converted, wait_on=wait_on)

    @background
    async def wait(self, value, eps=None, sleep_time=1e-1 * q.s, timeout=None):
        """Wait until the parameter value is *value*. *eps* is the allowed discrepancy between the
        actual value and *value*. *sleep_time* is the time to sleep between consecutive checks.
        *timeout* specifies the maximum waiting time.
        """
        async def condition():
            """Take care of rountind errors"""
            diff = await self.get() - value
            if eps is not None:
                diff = np.abs(diff.to(eps.units))
                if diff < eps:
                    diff = 0

            return diff == 0

        await wait_until(condition, sleep_time=sleep_time, timeout=timeout)

    def _check_limit(self, value):
        """Common tasks for lower and upper before we set them."""
        if self._limits_locked:
            raise LockError('upper limit locked')
        if not self.unit.is_compatible_with(value):
            raise UnitError("limit units must be compatible with `{}'".format(self.unit))


class SelectionValue(ParameterValue):

    """Descriptor for :class:`.Selection` class."""

    def __init__(self, instance, selection):
        super(SelectionValue, self).__init__(instance, selection)

    @background
    async def set(self, value, wait_on=None):
        """
        Set concrete *value* on the object.

        If *wait_on* is not None, it must be an awaitable on which this method waits.
        """
        if value not in self._parameter.iterable:
            raise SelectionError(self._parameter.name, value, self._parameter.iterable)

        await super(SelectionValue, self).set(value, wait_on=wait_on)

    @property
    def values(self):
        """Selection values."""
        return tuple(self._parameter.iterable)


class Parameterizable(object):

    """
    Collection of parameters.

    For each class of type :class:`.Parameterizable`, :class:`.Parameter` can
    be set as class attributes ::

        class Device(Parameterizable):

            def get_something(self):
                return 'something'

            something = Parameter(get_something)

    There is a simple :class:`.Parameter` and a parameter which models a
    physical quantity :class:`.Quantity`.

    A :class:`.Parameterizable` is iterable and returns its parameters of type
    :class:`.ParameterValue` or its subclasses ::

        for param in device:
            print("name={}".format(param.name))

    To access a single name parameter object, you can use the ``[]`` operator::

        param = device['position']

    If the parameter name does not exist, a :class:`.ParameterError` is raised.

    Each parameter value is accessible as a property. If a device has a
    position it can be read and written with::

        param.position = 0 * q.mm
        print param.position
    """

    def __init__(self):
        if not hasattr(self, '_params'):
            self._params = {}

        for base in self.__class__.__mro__:
            for attr_name, attr_type in list(base.__dict__.items()):
                if isinstance(attr_type, Parameter):
                    attr_type.name = attr_name
                    self._install_parameter(attr_type)

    def __str__(self):
        return run_in_loop(self.info_table).get_string(sortby="Parameter")

    def __repr__(self):
        return '\n'.join([super(Parameterizable, self).__repr__(), str(self)])

    def __iter__(self):
        for param in sorted(self._params.values()):
            yield param

    def __getitem__(self, param):
        if param not in self._params:
            raise ParameterError(param)

        return self._params[param]

    def __contains__(self, key):
        return key in self._params

    @property
    async def info_table(self):
        from concert.session.utils import get_default_table

        table = get_default_table(["Parameter", "Value"])
        table.border = False

        for param in self:
            try:
                value = str(await param.get())
            except ReadAccessError:
                value = 'N/A'
            table.add_row([param.name, value])

        return table

    def install_parameters(self, params):
        """Install parameters at run-time.

        *params* is a dictionary mapping parameter names to :class:`.Parameter`
        objects.
        """
        for name, param in list(params.items()):
            param.name = name
            self._install_parameter(param)

        # Obtain the object-dot notation.
        merged_dict = dict(list(self.__class__.__dict__.items()) + list(params.items()))
        self.__class__ = type(self.__class__.__name__, self.__class__.__bases__, merged_dict)

    def _install_parameter(self, param):
        if isinstance(param, Quantity):
            value = QuantityValue(self, param)
        elif isinstance(param, Selection):
            value = SelectionValue(self, param)
        elif isinstance(param, State):
            value = StateValue(self, param)
        else:
            value = ParameterValue(self, param)

        self._params[param.name] = value

        getter_name = 'get_' + param.name
        setter_name = 'set_' + param.name
        if getter_name not in self.__class__.__dict__:
            def get_parameter(instance, wait_on=None):
                return instance[param.name].get(wait_on=wait_on)
            setattr(self.__class__, getter_name, get_parameter)
        if setter_name not in self.__class__.__dict__:
            def set_parameter(instance, parameter_value, wait_on=None):
                return instance[param.name].set(parameter_value, wait_on=wait_on)
            setattr(self.__class__, setter_name, set_parameter)

        if not hasattr(self, '_set_' + param.name):
            setattr(self, '_set_' + param.name, _setter_not_implemented)

        if not hasattr(self, '_get_' + param.name):
            setattr(self, '_get_' + param.name, _getter_not_implemented)

        if not hasattr(self, '_get_target_' + param.name):
            setattr(self, '_get_target_' + param.name, _getter_target_not_implemented)

    @background
    async def stash(self):
        """
        Save all writable parameters that can be restored with
        :meth:`.Parameterizable.restore`.

        The values are stored on a stacked, hence subsequent saved states can
        be restored one by one.
        """
        await asyncio.gather(*(param.stash() for param in self if param.writable))

    @background
    async def restore(self):
        """Restore all parameters saved with :meth:`.Parameterizable.stash`."""
        await asyncio.gather(*(param.restore() for param in self if param.writable))

    def lock(self, permanent=False):
        """Lock all the parameters for writing. If *permanent* is True, the
        parameters cannot be unlocked anymore.
        """
        for param in self:
            param.lock(permanent=permanent)

    def unlock(self):
        """Unlock all the parameters for writing."""
        for param in self:
            param.unlock()



from itertools import product
import numpy as np
from concert.quantities import q
from concert.tests import assert_almost_equal, TestCase
from concert.devices.motors.dummy import LinearMotor
from concert.processes.common import scan, ascan, dscan


def compare_sequences(first_sequence, second_sequence, assertion):
    for x, y in zip(first_sequence, second_sequence):
        assertion(x[0], y[0])
        assertion(x[1], y[1])


class TestScan(TestCase):

    def setUp(self):
        super(TestScan, self).setUp()
        self.motor = LinearMotor()

    async def feedback(self):
        return 1 * q.dimensionless

    async def test_ascan(self):

        async def run():
            scanned = []
            async for pair in ascan(self.motor['position'], 0 * q.mm, 10 * q.mm,
                                    5 * q.mm, self.feedback):
                scanned.append(pair)
            return scanned

        expected = [(0 * q.mm, 1 * q.dimensionless), (5 * q.mm, 1 * q.dimensionless)]

        scanned = await run()
        compare_sequences(expected, scanned, assert_almost_equal)

        # Second scan, values must be same
        scanned = await run()
        compare_sequences(expected, scanned, assert_almost_equal)

    async def test_dscan(self):
        async def run():
            scanned = []
            async for pair in dscan(self.motor['position'], 10 * q.mm, 5 * q.mm, self.feedback):
                scanned.append(pair)
            return scanned

        scanned = await run()
        expected = [(0 * q.mm, 1 * q.dimensionless), (5 * q.mm, 1 * q.dimensionless)]
        compare_sequences(expected, scanned, assert_almost_equal)

        # Second scan, x values must be different
        scanned = await run()
        expected = [(5 * q.mm, 1 * q.dimensionless), (10 * q.mm, 1 * q.dimensionless)]
        compare_sequences(expected, scanned, assert_almost_equal)

    async def test_scan(self):
        async def run():
            scanned = []
            async for pair in scan(self.motor['position'], np.arange(0, 10, 5) * q.mm,
                                   self.feedback):
                scanned.append(pair)
            return scanned

        scanned = await run()
        expected = [(0 * q.mm, 1 * q.dimensionless), (5 * q.mm, 1 * q.dimensionless)]
        compare_sequences(expected, scanned, assert_almost_equal)

    async def test_multiscan(self):
        """A 2D scan."""
        values_0 = np.arange(0, 10, 5) * q.mm
        values_1 = np.arange(20, 30, 5) * q.mm

        async def run():
            other = LinearMotor()
            scanned = []
            async for pair in scan((self.motor['position'], other['position']),
                                   (values_0, values_1),
                                   self.feedback):
                vec, res = pair
                scanned.append((vec[0], vec[1], res))
            return scanned

        scanned = await run()
        expected = list(product(values_0, values_1, [1 * q.dimensionless]))
        x, y, z = list(zip(*scanned))
        x_gt, y_gt, z_gt = list(zip(*expected))
        assert_almost_equal(x, x_gt)
        assert_almost_equal(y, y_gt)
        assert_almost_equal(z, z_gt)



"""
Dummy ring to test asychronous updates.
"""
import time
import threading
from concert.quantities import q
from concert.devices.storagerings import base


class StorageRing(base.StorageRing):

    """A storage ring dummy."""

    def __init__(self):
        super(StorageRing, self).__init__()
        self._lifetime = 10 * q.hour
        self._current = 100 * q.mA
        self._energy = 5 * q.MeV
        self._current_decay = 0.05 * q.mA / q.hour
        self._energy_decay = 0.05 * q.MeV / q.hour

        def update():
            """Test updates."""
            while True:
                # Yay, we invented a time machine \o/
                self._lifetime += 1 * q.hour
                time.sleep(5.0)

        self.monitor = threading.Thread(target=update)
        self.monitor.daemon = True
        self.monitor.start()

    async def _get_current(self):
        return self._current - self._lifetime * self._current_decay

    async def _get_energy(self):
        return self._energy - self._lifetime * self._energy_decay

    async def _get_lifetime(self):
        return self._lifetime



"""Experiment automation based on on-line data analysis."""
from concert.coroutines.base import background


class ClosedLoop(object):
    """An abstract feedback loop which acquires data, analyzes it on-line and provides feedback to
    the experiment. The data acquisition procedure is done iteratively until the result of some
    metric converges to a satisfactory value. Schematically, the class is doing the following in an
    iterative way::

        initialize -> measure -> compare -> OK -> success
                        ^            |
                        |           NOK
                        |            |
                        -- control <--

    """
    async def initialize(self):
        """Bring the experimental setup to some defined initial (reference) state."""
        pass

    async def measure(self):
        """Conduct a measurement from data acquisition to analysis."""
        pass

    async def control(self):
        """React on the result of a measurement."""
        pass

    async def compare(self):
        """Return True if the metric is satisfied, False otherwise. This is the decision making
        process.
        """
        raise NotImplementedError

    @background
    async def run(self, max_iterations=10):
        """
        run(self, max_iterations=10)

        Run the loop until the metric is satisfied, if we don't converge in *max_iterations* then
        the run is considered unsuccessful and False is returned, otherwise True.
        """
        await self.initialize()

        for i in range(max_iterations):
            await self.measure()
            if await self.compare():
                return True
            else:
                await self.control()

        return False


class DummyLoop(ClosedLoop):

    """A dummy optimization loop."""

    async def compare(self):
        return True



import asyncio
import numpy as np
from concert.coroutines.base import (async_generate, broadcast, feed_queue, run_in_executor,
                                     run_in_loop, run_in_loop_thread_blocking, start, wait_until,
                                     WaitError)
from concert.coroutines.filters import (absorptivity, flat_correct, average_images,
                                        downsize, stall, Timer)
from concert.coroutines.sinks import null, Result, Accumulate
from concert.quantities import q
from concert.tests import assert_almost_equal, TestCase


async def produce_frames(num_frames=3, shape=(2, 2)):
    for i in range(num_frames):
        yield np.ones(shape=shape) * (i + 1)


class TestCoroutines(TestCase):

    def setUp(self):
        super(TestCoroutines, self).setUp()
        self.data = None
        self.iteration = 0

    async def produce(self, num_items=5):
        for i in range(num_items):
            yield i
            self.iteration += 1

    async def consume(self, producer):
        async for data in producer:
            self.data = data

    async def test_broadcast(self):

        async def consume(producer):
            nonlocal data
            async for item in producer:
                data = item

        data = None
        coros = broadcast(self.produce(), self.consume, consume)
        await asyncio.gather(*coros)
        self.assertEqual(self.data, 4)
        self.assertEqual(data, 4)

    async def test_broadcast_consumer_break(self):

        async def consume(producer):
            nonlocal data
            async for item in producer:
                data = item
                break

        data = None
        coros = broadcast(self.produce(), self.consume, consume)
        await asyncio.gather(*coros)
        # First gets the whole stream, the breaking one just the item before break
        self.assertEqual(self.data, 4)
        self.assertEqual(data, 0)

    async def test_broadcast_cancel(self):
        produced = asyncio.Event()

        async def produce():
            for i in range(3):
                yield i
                produced.set()
                # Allow future.cancel() to kick in
                await asyncio.sleep(0)

        coros = broadcast(produce(), self.consume)
        future = asyncio.gather(*coros)
        await produced.wait()
        future.cancel()
        await asyncio.wait([future])
        self.assertEqual(self.data, 0)

    async def test_feed_queue(self):
        produced = asyncio.Event()
        item = None

        async def produce():
            for i in range(3):
                yield i
                produced.set()
                # Allow future.cancel() to kick in
                await asyncio.sleep(0)

        def blocking_consumer(queue, check_cancelled=False):
            nonlocal item
            while True:
                item = queue.get()
                # print(f'{item.data}:{item.priority}')
                queue.task_done()
                if item.data is None:
                    break

        future = start(feed_queue(produce(), blocking_consumer))
        await future
        # Normal operation, priority with counter must arrive with None as data
        self.assertEqual(item.priority, 4)
        self.assertEqual(item.data, None)
        produced.clear()

        future = start(feed_queue(produce(), blocking_consumer))
        await produced.wait()
        future.cancel()
        # TODO: for some reason simple "await future" hangs forever, perhaps a bug in the test
        # suite?
        await asyncio.wait([future])
        # On cancel, the highest priority (0) must arrive with None as data
        self.assertTrue(future.cancelled())
        self.assertEqual(item.priority, 0)
        self.assertEqual(item.data, None)

    async def test_wait_until(self):
        ran = False

        async def do_slowly():
            nonlocal ran
            await asyncio.sleep(0.1)
            ran = True

        async def condition():
            return ran

        await asyncio.gather(wait_until(condition), do_slowly())
        self.assertTrue(ran)
        ran = False

        await asyncio.gather(wait_until(condition, timeout=10 * q.s), do_slowly())
        self.assertTrue(ran)
        ran = False

        with self.assertRaises(WaitError):
            await asyncio.gather(wait_until(condition, timeout=0.001 * q.s), do_slowly())

    def test_run_in_loop(self):
        async def consume():
            return 1

        self.assertEqual(run_in_loop(consume()), 1)

    async def test_run_in_loop_thread_blocking(self):
        async def process():
            return 1

        def blocking():
            return run_in_loop_thread_blocking(process())

        async def run_async_code_in_blocking():
            return blocking()

        self.assertEqual(await run_async_code_in_blocking(), 1)

    async def test_run_in_executor(self):
        def blocking(arg):
            return arg

        self.assertEqual(await run_in_executor(blocking, 'foo'), 'foo')

    async def test_start(self):
        future = start(null(self.produce()))
        self.assertTrue(isinstance(future, asyncio.Future))
        await asyncio.sleep(.1)
        self.assertTrue(self.iteration != 0)

    async def test_null(self):
        await null(self.produce())
        self.assertEqual(5, self.iteration)

    async def test_averager(self):
        await self.consume(average_images(produce_frames()))
        truth = np.ones((2, 2)) * 2
        np.testing.assert_almost_equal(self.data, truth)

    async def test_flat_correct(self):
        shape = (2, 2)
        dark = np.ones(shape)
        flat = np.ones(shape) * 10
        truth_base = np.ones(shape)

        async def check(producer):
            i = 1
            async for frame in producer:
                value = (i - 1) / 9.0
                np.testing.assert_almost_equal(frame, truth_base * value)
                i += 1

        await check(flat_correct(flat, produce_frames(), dark=dark))

    async def test_absorptivity(self):
        truth_base = np.ones((2, 2))

        async def check(producer):
            i = 1
            async for frame in producer:
                np.testing.assert_almost_equal(frame, -np.log(truth_base * i))
                i += 1
        await check(absorptivity(produce_frames()))

    async def test_result_object(self):
        result = Result()
        await result(self.produce())
        self.assertEqual(result.result, 4)

    async def test_downsize(self):
        acc = Accumulate()
        await acc(downsize(produce_frames(num_frames=10, shape=(10, 10)),
                           x_slice=(2, 6, 2), y_slice=(3, 10, 3), z_slice=(2, 8, 3)))
        result = np.array(acc.items)
        ones = np.ones(shape=(3, 2))

        self.assertEqual(result.shape, (2, 3, 2))
        np.testing.assert_almost_equal(result[0], ones * 3)
        np.testing.assert_almost_equal(result[1], ones * 6)

    async def test_stall(self):
        # Simple case, no modulo
        acc = Accumulate()
        await acc(stall(self.produce(num_items=10), per_shot=5))
        self.assertEqual(acc.items, [4, 9])

        # More iterations than total items
        await acc(stall(self.produce(num_items=26), per_shot=8, flush_at=13))
        self.assertEqual(acc.items, [7, 12, 20, 25])

    async def test_accumulate(self):
        async def run_test(data, shape=None, dtype=None):
            accumulate = Accumulate(shape=shape, dtype=dtype)
            await accumulate(async_generate(data))
            np.testing.assert_equal(accumulate.items, data)
            target_type = np.ndarray if shape else list
            self.assertTrue(isinstance(accumulate.items, target_type))
            if shape:
                self.assertEqual(accumulate.items.dtype, data.dtype)

        shape = (4, 4, 4)
        dtype = np.ushort
        data = np.ones(shape, dtype=dtype)
        await run_test(data)
        await run_test(data, shape=shape, dtype=dtype)


class TestTimer(TestCase):

    def setUp(self):
        self.timer = Timer()

    async def asyncSetUp(self):
        await null(self.timer(async_generate([1, 2, 3])))

    def test_durations(self):
        self.assertEqual(len(self.timer.durations), 3)

    def test_duration(self):
        """Test the sum of durations."""
        assert_almost_equal(self.timer.duration, sum(self.timer.durations))

    def test_mean(self):
        assert_almost_equal(self.timer.mean, self.timer.duration / len(self.timer.durations))

    def test_reset(self):
        self.timer.reset()
        self.assertEqual(len(self.timer.durations), 0)



from concert.devices.attenuatorboxes.dummy import AttenuatorBox
from concert.tests import TestCase


class TestAttenuatorBox(TestCase):

    def setUp(self):
        super(TestAttenuatorBox, self).setUp()
        self.attenuatorBox = AttenuatorBox()

    def test_set_attenuator(self):
        material0 = None
        self.attenuatorBox.attenuator = material0
        self.assertEqual(material0, self.attenuatorBox.attenuator)
        material1 = 'Al_1mm'
        self.attenuatorBox.attenuator = material1
        self.assertEqual(material1, self.attenuatorBox.attenuator)



from concert.devices.shutters.dummy import Shutter as DummyShutter
from concert.tests import TestCase


class TestDummyShutter(TestCase):

    def setUp(self):
        super(TestDummyShutter, self).setUp()
        self.shutter = DummyShutter()

    async def test_open(self):
        self.assertEqual(await self.shutter.get_state(), 'open')

    async def test_close(self):
        await self.shutter.close()
        self.assertEqual(await self.shutter.get_state(), 'closed')



from concert.base import TransitionNotAllowed
from concert.quantities import q
from concert.devices.xraytubes.dummy import XRayTube
from concert.tests import TestCase


class TestXrayTube(TestCase):

    def setUp(self):
        self.tube = XRayTube()

    async def test_on(self):
        if await self.tube.get_state() != 'off':
            await self.tube.off()

        await self.tube.on()
        self.assertEqual('on', await self.tube.get_state())
        with self.assertRaises(TransitionNotAllowed):
            await self.tube.on()

    async def test_off(self):
        if await self.tube.get_state() != 'on':
            await self.tube.on()

        await self.tube.off()
        self.assertEqual('off', await self.tube.get_state())
        with self.assertRaises(TransitionNotAllowed):
            await self.tube.off()

    def test_power(self):
        self.tube.current = 2 * q.A
        self.tube.voltage = 3 * q.V
        self.assertEqual(self.tube.power, 6 * q.W)



import numpy as np
from concert.quantities import q
from concert.devices.motors.dummy import ContinuousRotationMotor
from concert.imageprocessing import find_needle_tips, find_sphere_centers
from concert.processes.common import align_rotation_axis, ProcessError
from concert.tests import slow, TestCase
from concert.tests.util.rotationaxis import SimulationCamera


class TestDummyAlignment(TestCase):

    def setUp(self):
        super(TestDummyAlignment, self).setUp()
        self.x_motor = ContinuousRotationMotor()
        self.y_motor = ContinuousRotationMotor()
        self.z_motor = ContinuousRotationMotor()

        self.x_motor.position = 0 * q.deg
        self.y_motor.position = 0 * q.deg
        self.z_motor.position = 0 * q.deg

        self.camera = SimulationCamera(128, self.x_motor["position"],
                                       self.y_motor["position"],
                                       self.z_motor["position"])

        # Allow 1 px misalignment in y-direction.
        self.eps = np.arctan(2 / self.camera.rotation_radius) * q.rad

    async def align_check(self, x_angle, z_angle, has_x_motor=True, has_z_motor=True,
                          get_ellipse_points=find_needle_tips, get_ellipse_points_kwargs=None):
        """"Align and check the results."""
        await self.x_motor.set_position(z_angle)
        await self.z_motor.set_position(x_angle)

        x_motor = self.x_motor if has_x_motor else None
        z_motor = self.z_motor if has_z_motor else None

        await align_rotation_axis(self.camera, self.y_motor, x_motor=x_motor, z_motor=z_motor,
                                  get_ellipse_points=get_ellipse_points,
                                  get_ellipse_points_kwargs=get_ellipse_points_kwargs)

        # In our case the best perfectly aligned position is when both
        # motors are in 0.
        if has_x_motor:
            assert np.abs(await self.x_motor.get_position()) < self.eps
        if has_z_motor:
            assert np.abs(await self.z_motor.get_position()) < self.eps

    async def test_no_motor(self):
        with self.assertRaises(ProcessError):
            await align_rotation_axis(self.camera, self.y_motor, x_motor=None, z_motor=None)

    @slow
    async def test_out_of_fov(self):
        async def get_ones():
            return np.ones((self.camera.size,
                            self.camera.size))

        self.camera._grab_real = get_ones

        with self.assertRaises(ProcessError) as ctx:
            await align_rotation_axis(self.camera, self.y_motor,
                                      x_motor=self.x_motor,
                                      z_motor=self.z_motor)

        self.assertEqual("Error finding reference points", str(ctx.exception))

    @slow
    async def test_not_offcentered(self):
        self.camera.rotation_radius = 0
        with self.assertRaises(ValueError) as ctx:
            await align_rotation_axis(self.camera, self.y_motor,
                                      x_motor=self.x_motor,
                                      z_motor=self.z_motor)

        self.assertEqual("Sample off-centering too "
                         + "small, enlarge rotation radius.",
                         str(ctx.exception))

    @slow
    async def test_no_x_axis(self):
        """Test the case when there is no x-axis motor available."""
        await self.align_check(17 * q.deg, 11 * q.deg, has_z_motor=False)

    @slow
    async def test_no_z_axis(self):
        """Test the case when there is no x-axis motor available."""
        await self.align_check(17 * q.deg, 11 * q.deg, has_x_motor=False)

    @slow
    async def test_not_misaligned(self):
        "Perfectly aligned rotation axis."
        await self.align_check(0 * q.deg, 0 * q.deg)

    @slow
    async def test_only_x(self):
        """Only misaligned laterally."""
        await self.align_check(-17 * q.deg, 0 * q.deg)

    @slow
    async def test_only_z(self):
        """Only misaligned in the beam direction."""
        await self.align_check(0 * q.deg, 11 * q.deg)

    @slow
    async def test_huge_x(self):
        self.camera.scale = (3, 0.25, 3)
        await self.align_check(60 * q.deg, 11 * q.deg)

    @slow
    async def test_huge_z(self):
        self.camera.scale = (3, 0.25, 3)
        await self.align_check(11 * q.deg, 60 * q.deg)

    @slow
    async def test_positive(self):
        await self.align_check(17 * q.deg, 11 * q.deg)

    @slow
    async def test_negative_positive(self):
        await self.align_check(-17 * q.deg, 11 * q.deg)

    @slow
    async def test_positive_negative(self):
        await self.align_check(17 * q.deg, -11 * q.deg)

    @slow
    async def test_negative(self):
        await self.align_check(-17 * q.deg, -11 * q.deg)

    @slow
    async def test_sphere(self):
        """Test sphere instead of needle."""
        self.camera.rotation_radius = self.camera.size / 2
        self.camera.scale = (.5, .5, .5)
        self.camera.y_position = 0
        self.eps = np.arctan(2 / self.camera.rotation_radius) * q.rad
        await self.align_check(-17 * q.deg, 11 * q.deg,
                               get_ellipse_points=find_sphere_centers)

    @slow
    async def test_kwargs(self):
        """Test keyword arguments passing. Use sphere simulation and segmentation for that."""
        self.camera.rotation_radius = self.camera.size / 2
        self.camera.scale = (.25, .25, .25)
        self.camera.y_position = 0
        self.eps = np.arctan(2 / self.camera.rotation_radius) * q.rad
        await self.align_check(-17 * q.deg, 11 * q.deg,
                               get_ellipse_points=find_sphere_centers,
                               get_ellipse_points_kwargs={'correlation_threshold': 0.9})



from concert.experiments.control import ClosedLoop
from concert.tests import TestCase


class VisitorLoop(ClosedLoop):

    """A visitor loop checks visitations of control loop nodes. On the first run, it simulates
    unsuccessful result, the second run is already a success.
    """

    def __init__(self):
        super(VisitorLoop, self).__init__()
        self.initialized = False
        self.measured = False
        self.compared = False
        self.controlled = False
        self.iteration = 1

    async def initialize(self):
        self.initialized = True

    async def measure(self):
        self.measured = True

    async def control(self):
        self.controlled = True

    async def compare(self):
        self.compared = True
        self.iteration += 1

        return self.iteration > 2


class TestOptimizationLoop(TestCase):
    def setUp(self):
        self.loop = VisitorLoop()

    async def test_all_called(self):
        await self.loop.run()
        self.assertTrue(self.loop.initialized)
        self.assertTrue(self.loop.measured)
        self.assertTrue(self.loop.controlled)

    async def test_run(self):
        self.assertFalse(await self.loop.run(max_iterations=1))
        self.assertTrue(await self.loop.run())



'''
Lens changer, like used in optics with the possibility to motorized switch objectives.
Getter and Setter for objective and the
current state must be provided by the device implementation.
'''

from concert.base import State, AccessorNotImplementedError
from concert.devices.base import Device


class LensChanger(Device):
    '''Lens changer base class.'''
    state = State(default='standby')

    def __init__(self):
        super(LensChanger, self).__init__()

    async def _set_objective(self, objective):
        raise AccessorNotImplementedError

    async def _get_objective(self):
        raise AccessorNotImplementedError


class LensChangerError(Exception):
    ''' LensChanger related error'''
    pass



"""Test sessions."""
import asyncio
from threading import Thread, Event
from concert.coroutines.base import start, run_in_loop_thread_blocking
from concert.quantities import q
from concert.session.utils import abort_awaiting, check_emergency_stop
from concert.tests import slow, suppress_logging, TestCase


async def corofunc(inst):
    inst.started.set()
    try:
        await asyncio.sleep(1)
        return 1
    except asyncio.CancelledError:
        inst.cancelled = True


@slow
class TestAborting(TestCase):

    def setUp(self):
        suppress_logging()
        self.cancelled = False
        self.started = Event()
        super().setUp()

    def test_abort_awaiting_background(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        task = start(corofunc(self))
        abort_awaiting(background=True)
        assert task.cancelled()

    def test_abort_awaiting_thread(self):
        result = None

        def run():
            nonlocal result

            result = run_in_loop_thread_blocking(corofunc(self))

        thread = Thread(target=run)
        thread.start()
        self.started.wait()
        abort_awaiting(background=True)
        thread.join()
        self.assertEqual(result, None)

    async def test_check_emergency_stop(self):
        class Callable:
            num_called = 0
            aborted = False

            def __call__(self):
                result = False
                if self.num_called == 2:
                    self.aborted = True
                    result = True

                self.num_called += 1

                return result

        clb = Callable()
        cancelled_task = start(corofunc(self))
        task = start(check_emergency_stop(clb, poll_interval=1 * q.ms))

        await asyncio.sleep(.1)

        await cancelled_task
        assert self.cancelled
        assert clb.aborted
        # This means callable has been called after the flag has been set, enabling it to wait for
        # clearing the flag
        assert clb.num_called > 3

        # After the flag has been cleared, no more aborting of newly created tasks
        running_task = start(corofunc(self))
        await asyncio.sleep(.1)
        await running_task
        assert running_task.result() == 1



import asyncio
import concert.config
import concurrent.futures
import functools
import queue
import logging
import time
from threading import Thread
from concert.config import AIODEBUG
from concert.helpers import PrioItem
from concert.quantities import q


LOG = logging.getLogger(__name__)
# Loop which runs in a separate thread
_TLOOP = None


async def async_generate(iterable):
    for item in iterable:
        yield item


def run_in_loop(coroutine):
    """Wrap *coroutine* into a `asyncio.Task`, run it in the current loop, block until it finishes
    and return the result.  On KeyboardInterrupt, the task is cancelled.
    """
    loop = asyncio.get_event_loop()
    task = asyncio.ensure_future(coroutine, loop=loop)

    try:
        result = loop.run_until_complete(task)
        return result
    except KeyboardInterrupt:
        LOG.log(AIODEBUG, "KeyboardInterrupt in `%s', cancelling", task.get_coro().__qualname__)
        task.cancel()


def run_in_loop_thread_blocking(coroutine):
    """Run *coroutine* in a new loop in a separate thread. Useful for third party functions, which
    are unaware of asyncio and we want to feed them with our asyncio code (like in the optimization
    module).
    """
    global _TLOOP

    if _TLOOP:
        raise RuntimeError('Another coroutine is already running in a separate thread. '
                           'Currently only one coroutine can run in a separate thread at a time.')

    def run():
        _TLOOP.run_forever()

    _TLOOP = asyncio.new_event_loop()
    thread = Thread(target=run, daemon=False)
    thread.start()

    try:
        future = asyncio.run_coroutine_threadsafe(coroutine, _TLOOP)
        return future.result()
    except concurrent.futures.CancelledError as e:
        LOG.log(AIODEBUG, 'run_in_loop_thread_blocking: %s cancelled', coroutine.__qualname__)
    finally:
        _TLOOP.call_soon_threadsafe(_TLOOP.stop)
        thread.join()
        _TLOOP.close()
        _TLOOP = None


def run_in_executor(func, *args):
    """Run a blocking function *func* with signature func(\*args) in an executor."""
    # Leave this here for now, if there are problems with the normal run_in_executor returning a
    # future then let's re-visit
    # """Run a blocking function *func* with signature func(*args) in an executor, wrap the future
    # into a real coroutine and return it.
    # """
    # async def make_coro():
    #     loop = asyncio.get_running_loop()
    #     future = loop.run_in_executor(None, func, *args)
    #
    #     try:
    #         return await future
    #     except asyncio.CancelledError:
    #         # Wait for the future to finish even if we are cancelled.
    #         LOG.log(AIODEBUG, f'{func.__name__} waiting for concurrent.Future')
    #         return await future
    # return make_coro()
    loop = asyncio.get_event_loop()

    return loop.run_in_executor(None, func, *args)


def start(coroutine):
    """Wrap *coroutine* into a task and start its execution right away. The returned task will also
    be cancellable by ctrl-k.
    """
    task = asyncio.ensure_future(coroutine)
    task._is_concert_task = True

    return task


def background(coroutine):
    """Same as :func:`.start`, just meant to be used as a decorator."""
    @functools.wraps(coroutine)
    def inner(*args, **kwargs):
        return start(coroutine(*args, **kwargs))

    return inner


async def ensure_coroutine(func, *args, **kwargs):
    """func(\*args, \*\*kwargs) returns an awaitable which is wrapped here into a real coroutine.
    This is useful for turuning futures from other libraries, like Tango, into real coroutines.
    """
    return await func(*args, **kwargs)


def broadcast(producer, *consumers):
    """
    broadcast(producer, *consumers)

    Feed *producer* to all *consumers*.
    """
    loop = asyncio.get_event_loop()
    next_val = loop.create_future()
    consumed = loop.create_future()
    stop = object()
    num_consumers = len(consumers)
    used = num_consumers

    async def produce():
        try:
            async for elem in producer:
                if not next_val.cancelled():
                    next_val.set_result(elem)
                await consumed
                if not num_consumers:
                    break
            if not next_val.cancelled():
                next_val.set_result(stop)
        except Exception as e:
            LOG.warn(f"Exception `{e}' in broadcast")
            # Make sure the duplicates stop even if there is an exception
            if not next_val.cancelled():
                next_val.set_result(stop)
            raise

    async def duplicate():
        nonlocal next_val, consumed, used, num_consumers
        try:
            while True:
                val = await next_val
                if val is stop:
                    return
                yield val
                used -= 1
                if not used:
                    if not consumed.cancelled():
                        consumed.set_result(None)
                        consumed = loop.create_future()
                        next_val = loop.create_future()
                    used = num_consumers
                else:
                    await consumed
        except Exception as e:
            print('broadcast duplicate cancelled', e)
            raise
        finally:
            # Our consumer called break in its async for, remove it from waiting list
            num_consumers -= 1
            used -= 1
            if not used:
                if not consumed.cancelled():
                    consumed.set_result(None)
                    consumed = loop.create_future()
                    next_val = loop.create_future()
                used = num_consumers

    started = [produce()]
    for consumer in consumers:
        started.append(consumer(duplicate()))

    return started


async def feed_queue(producer, func, *args):
    """Feed function *func* with items from *producer* in a separete thread. The signatute must be
    func(queue, \*args) where elements in the queue are instances of
    :class:`concert.helpers.PrioItem`.
    """
    loop = asyncio.get_running_loop()
    pqueue = queue.PriorityQueue()

    try:
        future = loop.run_in_executor(None, func, pqueue, *args)
        prio = 1
        async for item in producer:
            pqueue.put(PrioItem(priority=prio, data=item))
            prio += 1
    except (asyncio.CancelledError, KeyboardInterrupt) as e:
        LOG.log(concert.config.AIODEBUG, f'feed_queue cancelled by exception {type(e)}')
        # Highest priority, effectively cancel processing of everything in the queue
        prio = 0
        raise
    finally:
        LOG.log(concert.config.AIODEBUG, f'feed_queue finished with priority {prio}')
        pqueue.put(PrioItem(priority=prio, data=None))
        await future


async def wait_until(condition, sleep_time=1e-1 * q.s, timeout=None):
    """Wait until a callable *condition* returns True. *sleep_time* is the time to sleep
    between consecutive checks of *condition*. If *timeout* is given and the *condition* doesn't
    return True within the time specified by it a :class:`.WaitingError` is raised.
    """
    sleep_time = sleep_time.to(q.s).magnitude
    if timeout:
        start = time.time()
        timeout = timeout.to(q.s).magnitude

    while not await condition():
        if timeout and time.time() - start > timeout:
            raise WaitError('Waiting timed out')
        await asyncio.sleep(sleep_time)


class WaitError(Exception):

    """Raised on busy waiting timeouts"""

    pass



"""PCO cameras implementation."""

import asyncio
import time
from datetime import datetime
import numpy as np
from concert.coroutines.base import background
from concert.devices.cameras import base
from concert.devices.cameras.uca import Camera
from concert.quantities import q


class Pco(Camera):

    """Pco camera implemented by libuca."""

    def __init__(self):
        super().__init__('pco')

    async def stream(self):
        """
        stream()

        Grab frames continuously yield them. This is an async generator.
        """
        try:
            await self.set_acquire_mode(self.uca.enum_values.acquire_mode.AUTO)
            await self.set_storage_mode(self.uca.enum_values.storage_mode.RECORDER)
            await self.set_record_mode(self.uca.enum_values.record_mode.RING_BUFFER)
        except Exception:
            pass

        async for image in super().stream():
            yield image


class PCO4000(Pco):

    """PCO.4000 camera implementation."""

    _ALLOWED_DURING_RECORDING = ['trigger', '_trigger_real', '_last_grab_time',
                                 'grab', '_grab_real', '_record_shape', '_record_dtype',
                                 'uca', 'stop_recording', 'convert', 'state', '_get_state',
                                 '_state_value']

    def __init__(self):
        self._lock_access = False
        self._last_grab_time = None
        super().__init__()

    @background
    async def start_recording(self):
        super().start_recording()
        self._lock_access = True

    @background
    async def stop_recording(self):
        self._lock_access = False
        super().stop_recording()

    async def _grab_real(self, index=None):
        # For the PCO.4000 there must be a delay of at least 1.2 second
        # between consecutive grabs, otherwise it crashes. We provide
        # appropriate timeout here.
        current = time.time()
        if self._last_grab_time and current - self._last_grab_time < 1.2:
            await asyncio.sleep(1.2 - current + self._last_grab_time)
        result = await super()._grab_real(index=index)
        self._last_grab_time = time.time()

        return result

    def __getattribute__(self, name):
        if object.__getattribute__(self, '_lock_access') and name \
           not in PCO4000._ALLOWED_DURING_RECORDING:
            raise AttributeError("{} cannot be accessed during recording".format(name))
        return object.__getattribute__(self, name)


class Dimax(Pco, base.BufferedMixin):

    """A pco.dimax camera implementation."""

    @background
    async def start_recording(self):
        await super().start_recording()
        # By low frame rates the camera returns status that it is already recording, whereas it is
        # not. Waiting 1 frame time seems to help.
        time.sleep((1 / self.frame_rate).to(q.s).magnitude)

    async def _readout_real(self, num_frames=None):
        """Readout *num_frames* frames."""
        recorded_frames = await self.get_recorded_frames()

        if num_frames is None:
            num_frames = recorded_frames.magnitude

        if not 0 < num_frames <= recorded_frames:
            raise base.CameraError("Number of frames {} ".format(num_frames)
                                   + "must be more than zero and less than the recorded "
                                   + "number of frames {}".format(recorded_frames))

        try:
            self.uca.start_readout()

            for i in range(num_frames):
                yield await self.grab()
        finally:
            self.uca.stop_readout()


class Timestamp(object):

    """Read PCO's binary timestamp. *sequence* is an array of 14 unsigned short pixels."""

    def __init__(self, image):
        """Constructor."""
        if image.dtype != np.uint16:
            raise TypeError('Sequence must have type unsigned short int')
        if len(image.shape) != 2 or image.shape[1] < 14:
            raise ValueError('Image must be a 2D image and must be at least 14 pixels wide.')

        # figures holds the individual figures from BCD-coded numbers, e.g a year composed of 4
        # digits is stored in four separate entries
        figures = []
        for num in image[0, :14]:
            # 16 bits per pixel, 4-bit BCD in the last 8 bits -> 2 decimal digits
            figures.append(num >> 4 & 0xf)
            figures.append(num & 0xf)

        self._number = _concatenate_ints(figures[:8])

        try:
            year = _concatenate_ints(figures[8:12])
            month = _concatenate_ints(figures[12:14])
            day = _concatenate_ints(figures[14:16])
            hour = _concatenate_ints(figures[16:18])
            minute = _concatenate_ints(figures[18:20])
            sec = _concatenate_ints(figures[20:22])
            usec = _concatenate_ints(figures[22:])
            self._time = datetime(year, month, day, hour, minute, sec, usec)
        except ValueError:
            raise TimestampError('No valid timestamp found.')

    @property
    def number(self):
        """Image number in the sequence."""
        return self._number

    @property
    def time(self):
        """Date and time when the image was taken."""
        return self._time

    def __repr__(self):
        return 'Timestamp(number={}, time={})'.format(self.number, self.time)


def _concatenate_ints(sequence):
    """Construct a number from *sequence* of integers, e.g. [1, 2, 3] will be transformed to
    123.
    """
    return int(''.join([str(num) for num in sequence]))


class TimestampError(Exception):
    """Bad timestamp error."""
    pass



from concert.tests import TestCase
from concert.quantities import q
from concert.devices.motors.base import RotationMotor as BaseRotationMotor
from concert.devices.motors.dummy import RotationMotor


class ImproperlyImplemented(BaseRotationMotor):
    def __init__(self):
        super(ImproperlyImplemented, self).__init__()
        self._value = 0

    async def _get_position(self):
        return self._value

    async def _set_position(self, value):
        self._value = value


class BreakingMotor(ImproperlyImplemented):
    def __init__(self):
        super(BreakingMotor, self).__init__()

    async def _get_state(self):
        return 'standby'


class TestIssue209(TestCase):

    def test_method_not_implemented(self):
        """
        Although required by base.RotationMotor, the BreakingMotor does not
        implement _get_state but rather than telling us that, it says that the
        parameter cannot be written.
        """
        fancy = ImproperlyImplemented()

        with self.assertRaises(NotImplementedError):
            fancy.position = 20 * q.deg

    def test_shared_parameters(self):
        dummy = RotationMotor()
        fancy = BreakingMotor()

        dummy.position = 10 * q.deg
        self.assertEqual(dummy.position, 10 * q.deg)

        fancy.position = 20 * q.deg
        self.assertEqual(fancy.position, 20 * q.deg)



"""Dummy experiments."""

import os
import numpy as np
from concert.coroutines.base import run_in_executor
from concert.experiments.base import Acquisition, Experiment
from concert.devices.cameras.dummy import FileCamera
from concert.progressbar import wrap_iterable


class ImagingExperiment(Experiment):

    """
    A typical imaging experiment which consists of acquiring dark, flat and radiographic images, in
    this case zeros or random data.

    .. py:attribute:: num_darks

        Number of dark images (no beam, just dark current)

    .. py:attribute:: num_flats

        Number of flat images (beam present, sample not)

    .. py:attribute:: num_radios

        Number of radiographic images

    .. py:attribute:: shape

        Shape of the generated images (H x W) (default: 1024 x 1024)

    .. py:attribute:: random

        'off': use zeros
        'single': one random repeated in every iteration
        'multi': every iteration generates new random image

    .. py:attribute:: dtype

        Data type of the generated images (default: unsigned short)
    """

    def __init__(self, num_darks, num_flats, num_radios, shape=(1024, 1024), walker=None,
                 random=False, dtype=np.ushort, separate_scans=True, name_fmt='scan_{:>04}'):
        self.num_darks = num_darks
        self.num_flats = num_flats
        self.num_radios = num_radios
        self.shape = shape
        if random not in ['off', 'single', 'multi']:
            raise ValueError("random must be one of 'off', 'single', 'multi'")
        self.random = random
        self.dtype = dtype
        darks = Acquisition('darks', self.take_darks)
        flats = Acquisition('flats', self.take_flats)
        radios = Acquisition('radios', self.take_radios)
        super(ImagingExperiment, self).__init__([darks, flats, radios], walker=walker)

    async def _produce_images(self, num):
        def make_random_image():
            return np.random.normal(128., 10., size=self.shape).astype(self.dtype)

        if self.random == 'off':
            image = await run_in_executor(np.zeros, self.shape, self.dtype)
        elif self.random == 'single':
            image = await run_in_executor(make_random_image)

        for i in wrap_iterable(list(range(num))):
            if self.random == 'multi':
                image = await run_in_executor(make_random_image)
            yield image

    def take_darks(self):
        return self._produce_images(self.num_darks)

    def take_flats(self):
        return self._produce_images(self.num_flats)

    def take_radios(self):
        return self._produce_images(self.num_radios)


class ImagingFileExperiment(Experiment):

    """
    A typical imaging experiment which consists of acquiring dark, flat and radiographic images, in
    this case located on a disk.

    .. py:attribute:: directory

       Top directory with subdirectories containing the individual images

    .. py:attribute:: num_darks

        Number of dark images (no beam, just dark current)

    .. py:attribute:: num_flats

        Number of flat images (beam present, sample not)

    .. py:attribute:: num_radios

        Number of radiographic images

    .. py:attribute:: darks_dir

        Subdirectory name with dark images

    .. py:attribute:: flats_dir

        Subdirectory name with flat images

    .. py:attribute:: radio_dir

        Subdirectory name with radiographic images

    .. py:attribute:: roi_x0

        First read column

    .. py:attribute:: roi_width

        Number of read columns

    .. py:attribute:: roi_y0

        First read row

    .. py:attribute:: roi_height

        Number of read rows
    """

    def __init__(self, directory, num_darks, num_flats, num_radios, darks_pattern='darks',
                 flats_pattern='flats', radios_pattern='projections', roi_x0=None, roi_width=None,
                 roi_y0=None, roi_height=None, walker=None, separate_scans=True,
                 name_fmt='scan_{:>04}'):
        self.directory = directory
        self.num_darks = num_darks
        self.num_flats = num_flats
        self.num_radios = num_radios
        self.darks_pattern = darks_pattern
        self.flats_pattern = flats_pattern
        self.radios_pattern = radios_pattern
        self.roi_x0 = roi_x0
        self.roi_width = roi_width
        self.roi_y0 = roi_y0
        self.roi_height = roi_height
        darks = Acquisition('darks', self.take_darks)
        flats = Acquisition('flats', self.take_flats)
        radios = Acquisition('radios', self.take_radios)
        super(ImagingFileExperiment, self).__init__([darks, flats, radios], walker=walker)

    async def _produce_images(self, pattern, num):
        camera = FileCamera(os.path.join(self.directory, pattern))
        if self.roi_x0 is not None:
            await camera.set_roi_x0(self.roi_x0)
        if self.roi_width is not None:
            await camera.set_roi_width(self.roi_width)
        if self.roi_y0 is not None:
            await camera.set_roi_y0(self.roi_y0)
        if self.roi_height is not None:
            await camera.set_roi_height(self.roi_height)

        for i in wrap_iterable(list(range(num))):
            yield await camera.grab()

    def take_darks(self):
        return self._produce_images(self.darks_pattern, self.num_darks)

    def take_flats(self):
        return self._produce_images(self.flats_pattern, self.num_flats)

    def take_radios(self):
        return self._produce_images(self.radios_pattern, self.num_radios)



"""Dummy IO"""
from concert.base import transition
from concert.devices.io import base


class IO(base.IO):

    """Dummy I/O device implementation."""

    def __init__(self, port_value=0):
        super(IO, self).__init__()
        self._ports = {0: port_value}

    async def _read_port(self, port):
        return self._ports[port]

    async def _write_port(self, port, value):
        self._ports[port] = value


class Signal(base.Signal):

    """Dummy signal device."""

    @transition(target='on')
    async def _on(self):
        pass

    @transition(target='off')
    async def _off(self):
        pass




class ImageWriter:

    def __init__(self, remote):
        self._remote = remote

    async def write_sequence(self, walker, acquisition_name):
        """Wrap the walker and write data."""
        async with walker:
            try:
                walker.descend(acquisition_name)
                await self._remote.start(walker)
            finally:
                walker.ascend()

    async def wait(self, count):
        await self._remote.wait(count)

    # def _teardown(self):
    #     self._remote.




from nose.plugins.attrib import attr
from concert.tests import TestCase
from concert.quantities import q


class TestIssue367(TestCase):
    @attr("skip-travis")
    def test_degree_conversion(self):
        from concert.devices.cameras.uca import Camera
        camera = Camera("mock")

        camera.degree_value = q.Quantity(5, q.celsius)
        self.assertEqual(camera.degree_value.magnitude, 5.0)

        val = camera.degree_value.magnitude
        camera.degree_value = camera.degree_value + 5 * q.delta_degC
        self.assertEqual(camera.degree_value.magnitude, val + 5)



"""Shutter Device."""
from concert.base import check, State, AccessorNotImplementedError
from concert.coroutines.base import background
from concert.devices.base import Device


class Shutter(Device):

    """Shutter device class implementation."""

    state = State(default="open")

    def __init__(self):
        super(Shutter, self).__init__()

    @background
    @check(source='closed', target='open')
    async def open(self):
        """open()

        Open the shutter."""
        await self._open()

    @background
    @check(source='open', target='closed')
    async def close(self):
        """close()

        Close the shutter."""
        await self._close()

    async def _open(self):
        raise AccessorNotImplementedError

    async def _close(self):
        raise AccessorNotImplementedError

    async def _emergency_stop(self):
        if await self.get_state() == 'open':
            await self.close()



import asyncio
import time
import numpy as np
from concert.coroutines.base import start, WaitError
from concert.quantities import q
from concert.tests import TestCase
from concert.base import (Parameterizable, Parameter, Quantity, State, transition, check,
                          SelectionError, SoftLimitError, LockError, ParameterError, UnitError,
                          WriteAccessError)
from concert.devices.dummy import SelectionDevice


class BaseDevice(Parameterizable):

    def __init__(self):
        super(BaseDevice, self).__init__()


async def _test_setter(device, value):
    device._test_value = value


async def _test_getter(device):
    return device._test_value


async def _test_target(device):
    return 10 * q.mm


class FooDevice(BaseDevice):

    state = State(default='standby')

    no_write = Parameter()
    foo = Quantity(q.m, check=check(source='*', target='moved'))
    bar = Quantity(q.m)
    test = Quantity(q.m, fset=_test_setter, fget=_test_getter)

    def __init__(self, default):
        super(FooDevice, self).__init__()
        self._value = default
        self._param_value = 0 * q.mm
        self._test_value = 0 * q.mm

    async def _get_foo(self):
        return self._value

    @transition(target='moved')
    async def _set_foo(self, value):
        self._value = value

    async def _get_bar(self):
        return 5 * q.m

    param = Parameter(fget=_get_foo, fset=_set_foo)

    async def _get_param(self):
        return self._param_value

    async def _set_param(self, value):
        self._param_value = value

    param = Parameter(fget=_get_param, fset=_set_param)


class FooDeviceTargetValue(BaseDevice):
    foo = Quantity(q.mm)

    def __init__(self, value):
        super(BaseDevice, self).__init__()
        self._value = value
        self._test_value = value

    async def _get_foo(self):
        delta = 1 * q.mm
        return self._value + delta

    async def _set_foo(self, value):
        self._value = value

    async def _get_target_foo(self):
        return self._value

    def test_setter(self, value):
        self._test_value = value

    def test_getter(self):
        return self._test_value

    def test_target(self):
        return 10 * q.mm

    test = Quantity(q.mm, fset=_test_setter, fget=_test_getter, fget_target=_test_target)


class FooDeviceTagetValue(BaseDevice):
    foo = Quantity(q.mm)
    test = Quantity(q.mm, fset=_test_setter, fget=_test_getter, fget_target=_test_target)

    def __init__(self, value):
        super(BaseDevice, self).__init__()
        self._value = value

    def _get_foo(self):
        delta = np.random.randint(low=1, high=100) * q.mm
        return self._value + delta

    def _set_foo(self, value):
        self._value = value

    def _get_target_foo(self):
        return self._value


class RestrictedFooDevice(FooDevice):
    def __init__(self, lower, upper):
        super(RestrictedFooDevice, self).__init__(0 * q.mm)
        self['foo'].lower = lower
        self['foo'].upper = upper


class ExternalLimitDevice(BaseDevice):

    foo = Quantity(q.mm,
                   external_lower_getter=lambda: -5 * q.mm,
                   external_upper_getter=lambda: 5 * q.mm)

    def __init__(self, value):
        super(ExternalLimitDevice, self).__init__()
        self._value = value


class AccessorCheckDevice(Parameterizable):

    foo = Quantity(q.m)

    def __init__(self, future, check):
        super(AccessorCheckDevice, self).__init__()
        self.check = check
        self.future = future
        self._value = 0 * q.mm

    async def _set_foo(self, value):
        self.check(self.future)
        self._value = value
        time.sleep(0.01)

    async def _get_foo(self):
        self.check(self.future)
        time.sleep(0.01)
        return self._value


class TestDescriptor(TestCase):

    def setUp(self):
        super(TestDescriptor, self).setUp()
        self.foo1 = FooDevice(42 * q.m)
        self.foo2 = FooDevice(23 * q.m)

    def test_property_identity(self):
        self.foo1.foo = 15 * q.m
        self.assertEqual(self.foo1.foo, 15 * q.m)
        self.assertEqual(self.foo2.foo, 23 * q.m)

    async def test_func_identity(self):
        await self.foo1.set_foo(15 * q.m)
        self.assertEqual(await self.foo1.get_foo(), 15 * q.m)
        self.assertEqual(await self.foo2.get_foo(), 23 * q.m)

    async def test_wait(self):
        await self.foo1.set_foo(1 * q.m)
        await self.foo1['foo'].wait(1 * q.m, eps=1e-3 * q.m)

        # No change on the parameter, timeout must take place
        with self.assertRaises(WaitError):
            await self.foo1['foo'].wait(0 * q.m, timeout=1e-5 * q.s)


class TestParameterizable(TestCase):

    def setUp(self):
        super(TestParameterizable, self).setUp()
        self.device = FooDevice(0 * q.mm)

    async def test_param(self):
        await self.device.set_param(15)
        self.assertEqual(await self.device.get_param(), 15)

    def test_is_writable(self):
        self.assertTrue(self.device['foo'].writable)
        self.assertFalse(self.device['no_write'].writable)
        self.assertTrue(self.device['param'].writable)

        with self.assertRaises(WriteAccessError):
            self.device.no_write = 42

    async def test_saving(self):
        await self.device.set_foo(1 * q.mm)
        await self.device.stash()
        await self.device.set_foo(2 * q.mm)
        await self.device.stash()
        await self.device.set_foo(0.123 * q.mm)
        await self.device.set_foo(1.234 * q.mm)

        await self.device.restore()
        self.assertEqual(await self.device.get_foo(), 2 * q.mm)

        await self.device.restore()
        self.assertEqual(await self.device.get_foo(), 1 * q.mm)

    async def test_manual_lock(self):
        self.device['foo'].lock()
        self.assertTrue(self.device['foo'].locked)

        with self.assertRaises(LockError):
            await self.device.set_foo(1 * q.m)

        # This must pass
        await self.device.set_param(1 * q.m)

        # Unlock and test if writable
        self.device['foo'].unlock()
        self.assertFalse(self.device['foo'].locked)
        await self.device.set_foo(1 * q.m)

        # Lock the whole device
        self.device.lock()

        with self.assertRaises(LockError):
            await self.device.set_param(1 * q.m)

        with self.assertRaises(LockError):
            await self.device.set_foo(1 * q.m)

        # Unlock the whole device and test if writable
        self.device.unlock()
        await self.device.set_param(1 * q.m)
        await self.device.set_foo(1 * q.m)

    def test_permanent_lock(self):
        self.device['foo'].lock(permanent=True)
        self.assertTrue(self.device['foo'].locked)

        with self.assertRaises(LockError):
            self.device.foo = 1 * q.m

        with self.assertRaises(LockError):
            self.device['foo'].unlock()

    def test_permanent_device_lock(self):
        self.device.lock(permanent=True)

        with self.assertRaises(LockError):
            self.device.unlock()

    async def test_wait_on_future(self):
        def check(future):
            if future:
                self.assertTrue(future.done())

        d1 = AccessorCheckDevice(None, check)
        f1 = start(d1.set_foo(2 * q.mm))
        d2 = AccessorCheckDevice(f1, check)
        f2 = start(d2.set_foo(5 * q.mm, wait_on=f1))
        d3 = AccessorCheckDevice(f2, check)
        f3 = start(d3.get_foo(wait_on=f2))
        await asyncio.gather(f1, f2, f3)


class TestParameter(TestCase):

    async def test_saving(self):
        device = FooDevice(0 * q.mm)
        await device['foo'].stash()
        await device.set_foo(1 * q.mm)
        await device['foo'].restore()
        self.assertEqual(await device.get_foo(), 0 * q.mm)

        with self.assertRaises(ParameterError):
            await device['no_write'].stash()

        with self.assertRaises(ParameterError):
            await device['no_write'].restore()

    async def test_saving_with_target_value(self):
        device = FooDeviceTargetValue(0 * q.mm)
        device['foo'].upper = None
        device['foo'].lower = None
        await device['foo'].stash()
        self.assertEqual(device['foo'].target_readable, True)
        await device.set_foo(1 * q.mm)
        self.assertEqual(device._value, 1 * q.mm)
        await device['foo'].restore()
        self.assertEqual(device._value, 0 * q.mm)

    def test_readonly_value(self):
        device = FooDevice(0 * q.mm)
        self.assertEqual(device.bar, 5 * q.m)
        self.assertEqual(device['bar'].writable, False)
        with self.assertRaises(WriteAccessError):
            device.bar = 1 * q.m

    def test_setter_getter_from_constructor(self):
        device = FooDevice(0 * q.mm)
        device.test = 1 * q.mm
        self.assertEqual(device['test'].writable, True)
        self.assertEqual(device.test, 1 * q.mm)

    def test_setter_getter_from_constructor_target(self):
        device = FooDeviceTargetValue(0 * q.mm)
        device.test = 1 * q.mm
        self.assertEqual(device.test, 1 * q.mm)
        self.assertEqual(device['test'].target_readable, True)
        self.assertEqual(device['test'].target, 10 * q.mm)

    def test_name_for_log(self):
        device = FooDevice(0 * q.mm)
        device.foo = 1 * q.mm
        self.assertEqual(device.name_for_log, 'device')


class TestQuantity(TestCase):

    def test_soft_limit_change(self):
        limited = FooDevice(0 * q.mm)
        limited['foo'].lower = -2 * q.mm
        limited['foo'].upper = 2 * q.mm

        limited.foo = -1.5 * q.mm
        limited.foo = +1.5 * q.mm

        with self.assertRaises(SoftLimitError):
            limited.foo = 2.5 * q.mm

    async def test_soft_limit_restriction(self):
        limited = RestrictedFooDevice(-2 * q.mm, 2 * q.mm)
        self.assertEqual(limited['foo'].lower, -2 * q.mm)
        self.assertEqual(limited['foo'].upper, +2 * q.mm)

    def test_setting_soft_limits_to_none(self):
        limited = RestrictedFooDevice(-2 * q.mm, 2 * q.mm)
        limited['foo'].upper = None
        limited.foo = 3 * q.mm
        limited['foo'].lower = None
        limited.foo = -3 * q.mm

    def test_parameter_property(self):
        device = FooDevice(42 * q.m)
        self.assertEqual(device['foo'].unit, q.m)

        with self.assertRaises(UnitError):
            device.foo = 2 * q.s

    def test_limits_lock(self):
        device = FooDevice(10 * q.mm)
        device['foo'].lock_limits()
        with self.assertRaises(LockError):
            device['foo'].lower = -10 * q.mm

        device['foo'].lock_limits(True)
        with self.assertRaises(LockError):
            device['foo'].unlock_limits()

    def test_limits_setting(self):
        dev = FooDevice(0 * q.mm)
        dev['foo'].lower = -1 * q.m
        dev['foo'].upper = 1 * q.um

        with self.assertRaises(UnitError):
            dev['foo'].lower = -1 * q.deg

        with self.assertRaises(UnitError):
            dev['foo'].upper = 1 * q.deg

    def test_limit_bounds(self):
        dev = FooDevice(0 * q.m)
        dev['foo'].lower = -1 * q.m
        dev['foo'].upper = 1 * q.m

        with self.assertRaises(ValueError):
            dev['foo'].lower = 2 * q.m

        with self.assertRaises(ValueError):
            dev['foo'].upper = -2 * q.m

    async def test_external_limits(self):
        dev = ExternalLimitDevice(0 * q.mm)
        self.assertEqual(dev['foo'].lower, -5 * q.mm)
        self.assertEqual(dev['foo'].upper, 5 * q.mm)

        dev['foo'].upper = 2 * q.mm
        self.assertEqual(dev['foo'].upper, 2 * q.mm)

        dev['foo'].upper = 10 * q.mm
        self.assertEqual(dev['foo'].upper, 5 * q.mm)

        dev['foo'].lower = -2 * q.mm
        self.assertEqual(dev['foo'].lower, -2 * q.mm)

        dev['foo'].lower = -10 * q.mm
        self.assertEqual(dev['foo'].lower, -5 * q.mm)


class TestSelection(TestCase):

    def setUp(self):
        self.device = SelectionDevice()

    def test_correct_access(self):
        for i in range(3):
            self.device.selection = i

    def test_wrong_access(self):
        with self.assertRaises(SelectionError):
            self.device.selection = 4

    def test_iterable_access(self):
        np.testing.assert_equal(self.device['selection'].values, list(range(3)))



from concert.devices.samplechangers.dummy import SampleChanger
from concert.tests import TestCase


class TestSampleChanger(TestCase):

    def setUp(self):
        super(TestSampleChanger, self).setUp()
        self.samplechanger = SampleChanger()

    def test_set_sample(self):
        self.samplechanger.sample = None
        self.assertEqual(None, self.samplechanger.sample)



"""
Detector is a device which consists of a camera and an objective lens
which affects the effective pixel size.
"""
from concert.quantities import q
from concert.base import Quantity, AccessorNotImplementedError
from concert.devices.base import Device


class Detector(Device):
    """
    A base detector class for holding cameras and optics necessary to
    do acquire image.
    """

    pixel_width = Quantity(q.m, help="Effective pixel width")
    pixel_height = Quantity(q.m, help="Effective pixel height")

    async def _get_pixel_width(self):
        return (await self.camera.get_sensor_pixel_width()) * self.magnification

    async def _get_pixel_height(self):
        return (await self.camera.get_sensor_pixel_height()) * self.magnification

    @property
    def camera(self):
        raise AccessorNotImplementedError

    @property
    def magnification(self):
        raise AccessorNotImplementedError



"""Opening images in external programs."""
import asyncio
import collections
import multiprocessing as mp
import os
import tempfile
import time
import logging
from queue import Empty
from subprocess import Popen
from typing import Callable
import numpy as np
from concert.base import Parameterizable, Parameter
from concert.coroutines.base import background, run_in_executor
from concert.quantities import q


LOG = logging.getLogger(__name__)
_MP_CTX = mp.get_context('spawn')


def imagej(image, path="imagej"):
    """
    imagej(image, path="imagej")

    Open *image* in ImageJ found by *path*.
    """
    def start_command(program, image):
        """
        Create a tmp file for dumping the *image* and use *program*
        to open that image. Use *writer* for writing the iamge to the disk.
        """
        from concert.storage import write_tiff

        tmp_file = tempfile.mkstemp()[1]
        try:
            full_path = write_tiff(tmp_file, image)
            with Popen([program, full_path]) as process:
                process.wait()
        finally:
            os.remove(full_path)
            LOG.debug('Temporary file removed')

    # Do not make a daemon process to make sure the interpreter waits for it to exit, i.e. the
    # *finally* will be called and temp file deleted.
    proc = _MP_CTX.Process(target=start_command, args=(path, image), daemon=False)
    proc.start()


class ViewerBase(Parameterizable):

    """
    A base class for data viewer which sends commands to a backend-specific updater which runs in a
    separate process.
    """

    def __init__(self):
        super().__init__()
        self._queue = _MP_CTX.Queue()
        # This prevents hanging in the case we exit the session after something is put in the queue
        # and before it is consumed.
        self._queue.cancel_join_thread()
        self._paused = False
        # To be set up by an actual implementation which runs the drawing backed in a separate
        # process
        self._proc = None
        # __del__ is not going to help because it's never called from our concert session

    @background
    async def __call__(self, producer, size=0, force=False):
        """
        Display stream from *producer*. If *size* is specified, stop after displaying *size* items.
        If *force* is True make sure the item is displayed, otherwise it may be skipped if there is
        something in the queue waiting to be shown.
        """
        i = 0

        async for item in producer:
            if not size or i < size:
                await self.show(item, force=force)

            i += 1

    @background
    async def show(self, item, force=False):
        """Push *item* to the queue for display in a separate proces. If *force* is True make sure
        the item is displayed, otherwise it may be skipped if there is something in the queue
        waiting to be shown.
        """
        # If the circumstances allow it, push the item to the queue for display
        # This must happen before instantiation of the updater below because _show may raise
        # exception, in which case we don't want the updater to have started yet.
        if not self._paused and (not self._queue.qsize() or force or not self._proc):
            await run_in_executor(self._show, item)

        # If there is no updater or it has been stopped, instantiate it and start it in a process.
        if not (self._proc and self._proc.is_alive()):
            updater = self._make_updater()
            self._proc = _MP_CTX.Process(target=updater.run, daemon=True)
            self._proc.start()
            # Make sure that all control commands, like changing colormap, have been processed
            while self._queue.qsize():
                await asyncio.sleep(0.01)

    def pause(self):
        """Pause, no images are dispayed but image commands work."""
        self._paused = True

    def resume(self):
        """Resume the viewer."""
        self._paused = False

    def _show(self, item):
        """Implementation of pushing *item* to the display queue."""
        raise NotImplementedError

    def _make_updater(self):
        """Updater factory method."""
        raise NotImplementedError


class PyplotViewer(ViewerBase):

    """
    Dynamic plot viewer using matplotlib.

    .. py:attribute:: style

        One of matplotlib's linestyle format strings

    .. py:attribute:: plt_kwargs

        Keyword arguments accepted by matplotlib's plot()

    .. py:attribute:: autoscale

        If True, the axes limits will be expanded as needed by the new data,
        otherwise the user needs to rescale the axes

    .. py:attribute:: title

        Plot title
    """

    style = Parameter(help='Line style')
    autoscale = Parameter(help='Autoscale view')

    def __init__(self, style: str = "o", plot_kwargs: dict = None, autoscale: bool = True,
                 title: str = ""):
        super().__init__()
        self._autoscale = autoscale
        self._style = style
        self._plot_kwargs = plot_kwargs
        self._title = title

    def _show(self, item):
        """Unravel the *item* for x and y so that it is plotted correctly."""
        try:
            if len(item) != 2:
                raise ValueError('Plotting accepts only (x, y) pairs')
        except TypeError as exc:
            raise ValueError('Plotting accepts only (x, y) pairs') from exc

        if isinstance(item, q.Quantity):
            item = item.magnitude
        first, second = item
        if isinstance(first, q.Quantity):
            first = first.magnitude
        if isinstance(second, q.Quantity):
            second = second.magnitude
        item = (first, second)

        self._queue.put(('plot', item))

    def _make_updater(self):
        return _PyplotUpdater(self._queue, self._style, self._plot_kwargs,
                              self._autoscale, title=self._title)

    def reset(self):
        """Clear the plotted data."""
        self._queue.put(('clear', None))

    async def _get_style(self):
        return self._style

    async def _set_style(self, style):
        """Set line style to *style*."""
        self._queue.put(('style', style))
        self._style = style

    async def _get_autoscale(self):
        return self._autoscale

    async def _set_autoscale(self, autoscale):
        """Set *autoscale* on the axes, can be True or False."""
        self._queue.put(('autoscale', autoscale))
        self._autoscale = autoscale


class ImageViewerBase(ViewerBase):

    """Backend-free base class for displaying images.

    .. py:attribute:: limits

        minimum and maximum gray value (black and white points). Can be a tuple (min, max), 'auto'
        or 'stream'. When 'auto', limits are adjusted for every shown image, when 'stream', limits
        are adjusted on every __call__.

    .. py:attribute:: downsampling

        Display every n-th pixel, which can speed up the viewer

    .. py:attribute:: title

        Image title

    .. py:attribute:: show_refresh_rate

        Whether or not to show refresh rate text directly embedded into the displayed image
    """

    show_refresh_rate = Parameter(help='Display current refresh rate')
    limits = Parameter(help='Black and white point')
    downsampling = Parameter(help='Display only every n-th pixel')

    def __init__(self, limits: str = 'stream', downsampling: int = 1, title: str = "",
                 show_refresh_rate: bool = False):
        super().__init__()
        self._show_refresh_rate = show_refresh_rate
        self._title = title
        self._downsampling = downsampling
        self._limits = limits

    @background
    async def __call__(self, producer: Callable, size: int = None, force: bool = False):
        # In case limits are set to 'stream' we need to reset clim
        self._queue.put(('clim', self._limits))
        return await super().__call__(producer, size=None, force=force)

    def _show(self, item):
        self._queue.put(('image', item[::self._downsampling, ::self._downsampling]))

    async def _get_downsampling(self):
        return self._downsampling

    async def _set_downsampling(self, value):
        if value not in range(1, 21):
            raise ValueError('Downsampling must be from interval [1, 20]')
        self._downsampling = value

    async def _get_limits(self):
        return self._limits

    async def _set_limits(self, limits):
        """
        Minimum and maximum gray value (black and white points). Can be a tuple (min, max), 'auto'
        or 'stream'. When 'auto', limits are adjusted for every shown image, when 'stream', limits
        are adjusted on every __call__.
        """
        if not (limits == 'auto' or limits == 'stream' or len(limits) == 2):
            raise ViewerError("limits can be a tuple (min, max), 'auto' or 'stream'")
        self._queue.put(('clim', limits))
        self._limits = limits

    async def _get_show_refresh_rate(self):
        return self._show_refresh_rate

    async def _set_show_refresh_rate(self, value):
        if not isinstance(value, bool):
            raise ValueError('boolean value expected')
        self._queue.put(('show-fps', value))
        self._show_refresh_rate = value


class PyQtGraphViewer(ImageViewerBase):

    """Dynamic image viewer using PyQtGraph."""

    def _make_updater(self):
        return _PyQtGraphUpdater(self._queue, limits=self._limits, title=self._title,
                                 show_refresh_rate=self._show_refresh_rate)


class PyplotImageViewer(ImageViewerBase):

    """Dynamic image viewer using matplotlib.

    .. py:attribute:: imshow_kwargs

        matplotlib's imshow keyword arguments

    .. py:attribute:: fast

        Whether to use the fast version without colorbar

    """

    colormap = Parameter(help='Colormap')

    def __init__(self, imshow_kwargs: dict = None, fast: bool = True, limits: str = 'stream',
                 downsampling: int = 1, title: str = "", show_refresh_rate: bool = False):
        super().__init__(limits=limits, downsampling=downsampling, title=title,
                         show_refresh_rate=show_refresh_rate)
        self._has_colorbar = not fast
        self._imshow_kwargs = {} if imshow_kwargs is None else imshow_kwargs
        self._make_imshow_defaults()

    def _make_updater(self):
        if self._has_colorbar:
            return _PyplotImageUpdater(
                self._queue, self._imshow_kwargs, limits=self._limits, title=self._title,
                show_refresh_rate=self._show_refresh_rate
            )
        return _SimplePyplotImageUpdater(
            self._queue, self._imshow_kwargs, limits=self._limits, title=self._title,
            show_refresh_rate=self._show_refresh_rate
        )

    def reset(self):
        """Reset the viewer's state."""
        self._queue.put(('reset', None))

    def _make_imshow_defaults(self):
        """Override matplotlib's image showing defafults."""
        if "cmap" not in self._imshow_kwargs:
            self._imshow_kwargs["cmap"] = 'gray'
        self._colormap = self._imshow_kwargs["cmap"]
        if "interpolation" not in self._imshow_kwargs:
            self._imshow_kwargs["interpolation"] = "nearest"

    async def _get_colormap(self):
        return self._colormap

    async def _set_colormap(self, colormap):
        """Set colormp of the shown image to *colormap*."""
        self._queue.put(('colormap', colormap))


class _PyQtGraphUpdater:

    """Fast PyQtGraph-based image viewing backend."""

    def __init__(self, queue: mp.Queue, limits: str = 'stream', title: str = "",
                 show_refresh_rate: bool = False):
        self.queue = queue
        self.title = title
        self.clim = limits
        self.show_refresh_rate = show_refresh_rate
        self.text = None
        self.plot = None
        # main graphics window
        self.view = None
        self.last_text_time = time.perf_counter()
        self.last_time = time.perf_counter()
        self.commands = {'image': self.proces_image,
                         'clim': self.update_limits,
                         'show-fps': self.toggle_show_refresh_rate}

    def process(self):
        """Process commands from queue."""
        try:
            cmd, item = self.queue.get(timeout=0.01)
            self.commands[cmd](item)
        except Empty:
            pass

    def update_all(self, image):
        """Display *image*."""
        now = time.perf_counter()
        self.view.imageItem.setImage(image, autoLevels=self.clim == 'auto', autoDownsample=True)
        if self.clim == 'stream':
            self.clim = (image.min(), image.max())
            self.sync_image_and_clim()

        if self.show_refresh_rate:
            if now - self.last_text_time > 0.5:
                fps = 1 / (now - self.last_time)
                self.text.setPos(image.shape[1] / 2, 0)
                self.text.setText(f'{fps:5.1f} FPS')
                self.last_text_time = now

        if not self.view.isVisible():
            self.view.show()

        self.last_time = now

    def proces_image(self, image):
        """Process current *image* including window setup if it is a first image."""
        import pyqtgraph as pg
        first = False

        if not self.view:
            first = True
            self.plot = pg.PlotItem(title=self.title)
            self.view = pg.ImageView(view=self.plot)
            self.make_refresh_rate_text()

        self.update_all(image)

        if first:
            self.update_limits(self.clim)

    def update_limits(self, clim):
        """Update limits (black and white point)."""
        # Store no matter if there is an image already or not
        self.clim = clim

        if not self.view or self.clim == 'stream':
            # No image has been displayed yet or the limits will be set with the next image
            return

        if clim == 'auto':
            # Synchronize histogram range and current image range drawn as lines
            self.view.imageItem.setImage(self.view.imageItem.image, autoLevels=True,
                                         autoDownsample=True)
            # Adjust histogram range
            self.view.ui.histogram.autoHistogramRange()
            return

        self.sync_image_and_clim()

    def sync_image_and_clim(self):
        self.view.imageItem.setLevels(self.clim)
        # Synchronize histogram range and current image range drawn as lines
        self.view.imageItem.setImage(self.view.imageItem.image, autoLevels=False,
                                     autoDownsample=True)
        self.view.ui.histogram.setHistogramRange(*self.clim)

    def toggle_show_refresh_rate(self, value):
        """Show refresh rate or not controlled by *value*."""
        self.show_refresh_rate = value

        if not self.view:
            return

        if value:
            self.make_refresh_rate_text()
        else:
            if self.text:
                self.view.removeItem(self.text)
            self.text = None

    def make_refresh_rate_text(self):
        """Make text for showing the refresh rate."""
        import pyqtgraph as pg

        if self.text or not self.show_refresh_rate:
            return

        self.text = pg.TextItem(anchor=(0.5, 0), color=(220, 220, 220), fill=(0, 0, 0))
        self.text.setOpacity(0.5)
        if self.view:
            self.view.addItem(self.text)

    def run(self):
        """Start drawing."""
        import pyqtgraph as pg
        from pyqtgraph.Qt import QtCore
        import signal

        signal.signal(signal.SIGINT, signal.SIG_IGN)

        app = pg.QtGui.QApplication([])
        # row-major for not transposing the array
        pg.setConfigOptions(antialias=True, imageAxisOrder='row-major')

        timer = QtCore.QTimer()
        timer.timeout.connect(self.process)
        # 0 ms delay -> be as fast as possible
        timer.start(0)

        app.exec_()


class _PyplotUpdaterBase:

    """
    Base class for animating a matploblib figure in a separate process.

    .. py:attribute:: queue

        A multiprocessing queue for receiving commands
    """

    def __init__(self, queue: mp.Queue, title: str = ""):
        self.queue = queue
        self.first = True
        self.title = title
        # A dictionary in form command: method which tells the class what to do
        # for every received command
        self.commands = {}

    def process(self, iteration):
        """Get item from queue and process it."""
        try:
            if self.first:
                # Wait as much time as it takes for the first
                # time beacuse we don't want to show a window
                # with no image in it.
                item = self.queue.get()
                self.first = False
            else:
                item = self.queue.get(timeout=0.01)
            cmd, data = item
            self.commands[cmd](data)
            return self.get_artists()
        except Empty:
            self.on_empty()
            return self.get_artists()

    def on_empty(self):
        """Callback on queue timeout."""

    def get_artists(self):
        """
        Abstract function for getting all matplotlib artists which we want
        to redraw. Needs to be implemented by the subclass.
        """
        raise NotImplementedError

    def run(self):
        """
        Run the process, i.e. wait for an image to come to the queue
        and dispaly it. This method is executed in a separate process.
        """
        # This import *must* be here, otherwise it doesn't work on Linux
        from matplotlib import pyplot as plt
        from matplotlib.animation import FuncAnimation
        # KeyboardInterrupt is ignored in order not to keep hanging somewhere...
        import signal
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        figure = plt.figure()
        animation = FuncAnimation(figure, self.process, interval=200, blit=False)
        plt.show()


class _PyplotUpdater(_PyplotUpdaterBase):

    """Plotting backend. The arguments are the same as by :py:class:`PyplotViewer`."""

    def __init__(self, queue: mp.Queue, style: str = "o", plot_kwargs: dict = None,
                 autoscale: bool = True, title: str = ""):
        super().__init__(queue, title=title)
        self.data = [[], []]
        self.line = None
        self.style = style
        self.plot_kwargs = {} if plot_kwargs is None else plot_kwargs
        if 'color' not in self.plot_kwargs:
            # Matplotlib changes colors all the time by default
            self.plot_kwargs['color'] = 'b'
        self.autoscale = autoscale
        self.commands = {'plot': self.plot,
                         'style': self.change_style,
                         'clear': self.clear,
                         'autoscale': self.set_autoscale}

    def get_artists(self):
        """Get the artists for the drawing."""
        return [] if self.line is None else [self.line]

    def make_line(self):
        """Create the line based on current settings."""
        from matplotlib import pyplot as plt

        self.line = plt.plot(self.data[0], self.data[1], self.style,
                             **self.plot_kwargs)[0]
        self.line.axes.set_title(self.title)

    def clear(self, data):
        """Clear everything from the plot."""
        self.data = [[], []]
        if self.line is not None:
            self.line.axes.clear()
        self.make_line()

    def plot(self, data):
        """Plot *data*, which is an (x, y) tuple."""
        def get_magnitude_and_unit(value):
            if hasattr(value, "magnitude"):
                dimless = value.magnitude
                unit = value.units
            else:
                dimless = value
                unit = None

            return dimless, unit

        x_item, x_units = get_magnitude_and_unit(data[0])
        y_item, y_units = get_magnitude_and_unit(data[1])

        first = len(self.data[0]) == 0
        is_iterable = isinstance(x_item, collections.Iterable)
        if is_iterable:
            self.data = [list(x_item), list(y_item)]
        else:
            self.data[0].append(x_item)
            self.data[1].append(y_item)

        if first:
            self.make_line()
            if not is_iterable:
                # Make sure the limits are set properly for the first time
                self.line.axes.set_xlim(self.data[0][0] - 1e-7, self.data[0][0] + 1e-7)
                self.line.axes.set_ylim(self.data[1][0] - 1e-7, self.data[1][0] + 1e-7)
            if x_units is not None:
                self.line.axes.get_xaxis().set_label_text(str(x_units))
            if y_units is not None:
                self.line.axes.get_yaxis().set_label_text(str(y_units))

        self.line.set_data(*self.data)

        if self.autoscale or first:
            # Draw for the first time or on limit change to update ticks and
            # labels
            self.autoscale_view()

    def change_style(self, style):
        """Change line style to *style*."""
        self.style = style
        if self.line is not None:
            # Just redrawing doesn't work
            self.line.axes.clear()
            self.make_line()

    def set_autoscale(self, autoscale):
        """If *autoscale* is True, the plit is rescaled when needed."""
        self.autoscale = autoscale
        if self.autoscale:
            self.autoscale_view()

    def autoscale_view(self):
        """Autoscale axes limits."""
        if self.line is not None:
            self.line.axes.relim()
            # For some reason the relim itself doesn't work, so we set the
            # limits to the new values explicitly
            if len(self.data[0]) > 1:
                self.line.axes.set_xlim(min(self.data[0]), max(self.data[0]))
                self.line.axes.set_ylim(min(self.data[1]), max(self.data[1]))
            self.line.axes.autoscale_view()


class _PyplotImageUpdaterBase(_PyplotUpdaterBase):

    """Common class for various image viewing backends."""

    def __init__(self, queue: mp.Queue, imshow_kwargs: dict, limits: str = 'stream',
                 title: str = "", show_refresh_rate: bool = False):
        super().__init__(queue, title=title)
        self.imshow_kwargs = imshow_kwargs
        self.mpl_image = None
        self.clim = limits
        self.text = None
        self.show_refresh_rate = show_refresh_rate
        self.last_text_time = time.perf_counter()
        self.last_time = time.perf_counter()
        self.commands = {'image': self.process_image,
                         'clim': self.update_limits,
                         'colormap': self.update_colormap,
                         'reset': self.reset,
                         'show-fps': self.toggle_show_refresh_rate}

    def process_image(self, image):
        """Display *image*."""
        if self.mpl_image is not None and self.mpl_image.get_size() != image.shape:
            self.reset()

        if self.mpl_image:
            # Either removed by shape change or first time drawing
            self.update_all(image)
        else:
            self.make_image(image)

    def make_image(self, image):
        """Setup everything and display *image* for the first time."""
        raise NotImplementedError

    def update_all(self, image):
        """Update everything which needs to be updated when new *image* arrives."""
        self.mpl_image.set_data(image)
        self.update_refresh_rate_text()
        if self.clim in ['auto', 'stream']:
            # If the limit is not set to a value we autoscale
            new_lower = float(image.min())
            new_upper = float(image.max())
            if self.limits_changed(new_lower, new_upper):
                self.mpl_image.set_clim(new_lower, new_upper)
            if self.clim == 'stream':
                self.clim = self.mpl_image.get_clim()

    def update_limits(self, clim):
        """Update limits (black and white point)."""
        # Store no matter if there is an image already or not
        self.clim = clim

        if self.mpl_image is None or clim == 'stream':
            return

        if clim == 'auto':
            image = self.mpl_image.get_array()
            clim = (image.min(), image.max())

        self.mpl_image.set_clim(clim)

    def update_colormap(self, colormap):
        """Update colormap."""
        raise NotImplementedError

    def reset(self, *args):
        """Reset state."""
        if self.mpl_image:
            self.mpl_image.remove()
            self.mpl_image = None
        if self.text:
            self.text.remove()
            self.text = None

    def toggle_show_refresh_rate(self, value):
        """Show refresh rate or not controlled by *value*."""
        self.show_refresh_rate = value

        if self.show_refresh_rate:
            if self.mpl_image:
                self.make_refresh_rate_text()
        else:
            if self.text:
                self.text.remove()
                self.text = None

    def get_artists(self):
        """Needed by matplotlib's FuncAnimation."""
        artists = []
        if self.mpl_image:
            artists.append(self.mpl_image)
        if self.text:
            artists.append(self.text)

        return artists

    def update_refresh_rate_text(self):
        """Update refresh rate text."""
        if not (self.show_refresh_rate and self.text):
            return

        current = time.perf_counter()
        if current - self.last_text_time > 0.5:
            # Don't update FPS text that often to be readable
            fps = 1 / (current - self.last_time)
            self.text.set_text(f'{fps:5.1f} FPS')
            self.last_text_time = current
        self.last_time = current

    def make_refresh_rate_text(self, animated=False):
        """Make text for showing the refresh rate."""
        import matplotlib.pyplot as plt

        if self.text or not self.show_refresh_rate:
            return

        self.text = plt.text(0.5, 0.975, '', alpha=0.5, animated=animated,
                             ha='center', va='center', backgroundcolor='black',
                             color='lightgray', transform=self.mpl_image.axes.transAxes)
        self.text.get_bbox_patch().set_alpha(0.5)

    def limits_changed(self, lower, upper):
        """
        Determine whether the colormap limits changed enough for colorbar
        redrawing.
        """
        if lower >= upper:
            return False
        new_range = upper - lower
        lower_ratio = np.abs(lower - self.mpl_image.get_clim()[0]) / new_range
        upper_ratio = np.abs(upper - self.mpl_image.get_clim()[1]) / new_range

        return lower_ratio > 0.1 or upper_ratio > 0.1


class _PyplotImageUpdater(_PyplotImageUpdaterBase):

    """Image viewing backend."""

    def __init__(self, queue: mp.Queue, imshow_kwargs: dict, limits: str = 'stream',
                 title: str = "", show_refresh_rate: bool = False):
        super().__init__(queue, imshow_kwargs, limits=limits, title=title,
                         show_refresh_rate=show_refresh_rate)
        self.colorbar = None

    def get_artists(self):
        """Get artists to return for matplotlib's animation."""
        artists = super().get_artists()
        if self.colorbar:
            artists.append(self.colorbar)

        return artists

    def update_limits(self, clim):
        super().update_limits(clim)
        if self.mpl_image:
            self.update_colorbar()

    def update_colormap(self, colormap):
        """Update colormap to *colormap*."""
        if self.mpl_image is not None:
            self.mpl_image.set_cmap(colormap)

        # Save for later
        self.imshow_kwargs["cmap"] = colormap

    def update_all(self, image):
        """Update image and colorbar."""
        super().update_all(image)
        self.update_colorbar()

    def reset(self, *args):
        # We need to get rid of the colorbar as well
        self.mpl_image.axes.figure.clear()
        self.mpl_image = None
        self.colorbar = None

    def update_colorbar(self):
        """Update the colorbar (rescale and redraw)."""
        shape = self.mpl_image.get_size()
        if (shape[1] > shape[0] and self.colorbar.orientation == 'vertical'
                or shape[0] >= shape[1] and self.colorbar.orientation == 'horizontal'):
            self.colorbar.remove()
            self.make_colorbar()

    def make_colorbar(self):
        """Make colorbar according to the current colormap."""
        from matplotlib import pyplot as plt

        colormap = self.imshow_kwargs.get("cmap")
        shape = self.mpl_image.get_size()
        orientation = 'horizontal' if shape[1] > shape[0] else 'vertical'
        self.colorbar = plt.colorbar(cmap=colormap, orientation=orientation)

    def make_image(self, image):
        """Create an image with colorbar"""
        from matplotlib import pyplot as plt
        self.mpl_image = plt.imshow(image, **self.imshow_kwargs)
        self.mpl_image.axes.set_title(self.title)
        self.make_refresh_rate_text(animated=False)
        if self.colorbar is None:
            self.make_colorbar()
        self.update_limits(self.clim)


class _SimplePyplotImageUpdater(_PyplotImageUpdaterBase):

    """Simple image viewing backend optimized for speed, no colorbar available."""

    def __init__(self, queue: mp.Queue, imshow_kwargs: dict, limits: str = 'stream',
                 title: str = "", show_refresh_rate: bool = False):
        super().__init__(queue, imshow_kwargs, limits=limits, title=title,
                         show_refresh_rate=show_refresh_rate)
        self.fig = None
        self.axes = None
        self.background = None
        self.closed = True

    def run(self):
        """There is no animation, we handle everything ourselves so just keep getting images from
        the queue.
        """
        # KeyboardInterrupt is ignored in order not to keep hanging somewhere...
        import signal
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        while True:
            self.process(None)

    def on_empty(self):
        """Let the image window be responsive and run the GUI event loop for a while.

        Note:
        plt.pause makes the window steal focus aggresively (mpl 3.1.2) and after minimizing it pops
        right back up. This keeps happening until user manipulates the image via the toolbar. Doing
        things manually via draw_idle and starting the event loop seems to do the trick, for now...
        """
        # plt.pause(0.01)
        if self.fig:
            self.fig.canvas.draw_idle()
            self.fig.canvas.start_event_loop(0.01)

    def on_close(self, event):
        """Called when the window is closed. Note it and re-open on next image."""
        self.closed = True

    def on_draw(self, event):
        """Called when size is changed and we need to copy the new bbox."""
        if event is not None:
            if event.canvas != self.fig.canvas:
                raise RuntimeError

        self.background = self.fig.canvas.copy_from_bbox(self.fig.bbox)
        if self.mpl_image:
            self.axes.draw_artist(self.mpl_image)
        if self.text:
            self.axes.draw_artist(self.text)

    def make_image(self, image):
        """Create a new matplotlib image."""
        import matplotlib.pyplot as plt

        self.mpl_image = self.axes.imshow(image, animated=True, **self.imshow_kwargs)
        if self.clim not in ['auto', 'stream']:
            self.mpl_image.set_clim(self.clim)
        self.make_refresh_rate_text(animated=True)
        # This makes sure axes BBox fits the newly displayed image
        self.axes.relim()
        # In case we were panning/zooming, this will reset the position
        self.axes.autoscale()
        # This will reset toolbar buttons
        self.axes.figure.canvas.toolbar.update()
        plt.show(block=False)
        plt.pause(0.1)
        self.background = self.fig.canvas.copy_from_bbox(self.fig.bbox)

    def update_limits(self, clim):
        super().update_limits(clim)
        if self.mpl_image:
            self.redraw()

    def update_colormap(self, colormap):
        """Update colormap to *colormap*."""
        if self.mpl_image is not None:
            self.mpl_image.set_cmap(colormap)
            self.redraw()

        # Save for after reset
        self.imshow_kwargs["cmap"] = colormap

    def update_all(self, image):
        self.fig.canvas.restore_region(self.background)
        super().update_all(image)
        self.redraw()

    def redraw(self):
        """Redraw scene."""
        self.axes.draw_artist(self.mpl_image)
        if self.text:
            self.axes.draw_artist(self.text)
        self.fig.canvas.blit(self.fig.bbox)
        self.fig.canvas.flush_events()

    def process_image(self, image):
        import matplotlib.pyplot as plt

        if self.closed:
            # Start from scratch after close or first invocation
            self.fig, self.axes = plt.subplots()
            self.mpl_image = None
            self.axes.set_title(self.title)
            self.fig.canvas.mpl_connect('draw_event', self.on_draw)
            self.fig.canvas.mpl_connect('close_event', self.on_close)
            self.closed = False

        super().process_image(image)


class ViewerError(Exception):
    """Viewer errors."""



"""This module provides a simple dummy camera."""
import asyncio
import time
import numpy as np
from concert.coroutines.base import run_in_executor
from concert.quantities import q
from concert.base import transition, Quantity
from concert.devices.cameras import base
from concert.readers import TiffSequenceReader


class Base(base.Camera):

    exposure_time = Quantity(q.s)
    sensor_pixel_width = Quantity(q.micrometer)
    sensor_pixel_height = Quantity(q.micrometer)
    roi_x0 = Quantity(q.pixel)
    roi_y0 = Quantity(q.pixel)
    roi_width = Quantity(q.pixel)
    roi_height = Quantity(q.pixel)

    def __init__(self):
        super(Base, self).__init__()
        self._frame_rate = 1000 / q.s
        self._trigger_source = self.trigger_sources.AUTO
        self._exposure_time = 1 * q.ms
        self._roi_x0 = 0 * q.pixel
        self._roi_y0 = 0 * q.pixel
        self._roi_width = 640 * q.pixel
        self._roi_height = 480 * q.pixel

    async def _get_sensor_pixel_width(self):
        return 5 * q.micrometer

    async def _get_sensor_pixel_height(self):
        return 5 * q.micrometer

    async def _get_roi_x0(self):
        return self._roi_x0

    async def _set_roi_x0(self, x0):
        self._roi_x0 = x0

    async def _get_roi_y0(self):
        return self._roi_y0

    async def _set_roi_y0(self, y0):
        self._roi_y0 = y0

    async def _get_roi_width(self):
        return self._roi_width

    async def _set_roi_width(self, roi):
        self._roi_width = roi

    async def _get_roi_height(self):
        return self._roi_height

    async def _set_roi_height(self, roi):
        self._roi_height = roi

    async def _get_exposure_time(self):
        return self._exposure_time

    async def _set_exposure_time(self, value):
        self._exposure_time = value

    async def _get_frame_rate(self):
        return self._frame_rate

    async def _set_frame_rate(self, frame_rate):
        self._frame_rate = frame_rate

    async def _get_trigger_source(self):
        return self._trigger_source

    async def _set_trigger_source(self, source):
        self._trigger_source = source

    @transition(target='recording')
    async def _record_real(self):
        pass

    @transition(target='standby')
    async def _stop_real(self):
        pass

    async def _trigger_real(self):
        pass


class Camera(Base):

    """A simple dummy camera."""

    def __init__(self, background=None, simulate=True):
        """
        *background* can be an array-like that will be used to generate the frame when calling
        :meth:`.grab`. If *simulate* is True the final image intensity will be scaled based on
        exposure time and poisson noise will be added. If *simulate* is False, the background will
        be returned with no modifications to it.
        """
        super(Camera, self).__init__()
        self.simulate = simulate

        if background is not None:
            self._roi_width = background.shape[1] * q.pixel
            self._roi_height = background.shape[0] * q.pixel
            self._background = background
        else:
            shape = (640, 480)
            self._roi_width, self._roi_height = shape * q.pixel
            self._background = np.ones(shape[::-1], dtype=np.uint8)

    async def _grab_real(self):
        if not self.simulate:
            return self._background
        start = time.time()
        cur_time = (await self.get_exposure_time()).to(q.s).magnitude
        # 1e5 is a dummy correlation between exposure time and emitted e-.
        tmp = self._background + cur_time * 1e5
        max_value = np.iinfo(np.uint16).max
        tmp = np.random.poisson(tmp)
        # Cut values beyond the bit-depth.
        tmp[tmp > max_value] = max_value
        duration = time.time() - start
        to_sleep = 1 / (await self.get_frame_rate())
        to_sleep = to_sleep.to_base_units() - duration * q.s
        if to_sleep > 0 * q.s:
            await asyncio.sleep(to_sleep.magnitude)

        return np.cast[np.uint16](tmp)


class FileCamera(Base):

    """A camera that reads files specified by *pattern*. It can be a directory, in which case all
    the files inside are read, or it can be a pattern and only the matching files will be read. If
    *reset_on_start* is True the files are read from the beginning when the recording starts.
    *start_index* specifies the index of the first read image (not file index, in case the files are
    multi-page).
    """

    def __init__(self, pattern, reset_on_start=True, start_index=0):
        # Let users change the directory
        super(FileCamera, self).__init__()
        self._reader = TiffSequenceReader(pattern)
        self.index = start_index
        self.reset_on_start = reset_on_start
        self._roi_height = 0 * q.px
        self._roi_width = 0 * q.px

    @transition(target='recording')
    async def _record_real(self):
        if self.reset_on_start:
            self.index = 0

    async def _grab_real(self):
        roi_x0 = await self.get_roi_x0()
        roi_y0 = await self.get_roi_y0()
        if await self.get_roi_height():
            y_region = (roi_y0 + await self.get_roi_height()).magnitude
        else:
            y_region = None
        if await self.get_roi_width():
            x_region = (roi_x0 + await self.get_roi_width()).magnitude
        else:
            x_region = None

        image = await run_in_executor(self._reader.read, self.index)
        self.index += 1

        return image[roi_y0.magnitude:y_region, roi_x0.magnitude:x_region]


class BufferedCamera(Camera, base.BufferedMixin):

    async def _readout_real(self):
        for i in range(3):
            yield await self.grab()



import logging
import time
import numpy as np
from concert.quantities import q
from concert.imageprocessing import flat_correct as make_flat_correct


LOG = logging.getLogger(__name__)


async def downsize(producer, x_slice=None, y_slice=None, z_slice=None):
    """
    downsize(producer, x_slice=None, y_slice=None, z_slice=None)

    Downsize images in 3D. Every argument is either a tuple (start, stop, step). *x_slice* operates
    on image width, *y_slice* on its height and *z_slice* on the incoming images, i.e. it creates
    the third time dimension.

    Note: the *start* index is included in the data and the *stop* index
    is excluded.
    """
    def check_and_create(sl):
        if not sl:
            sl = (0, None, 1)
        return sl

    x_start, x_stop, x_step = check_and_create(x_slice)
    y_start, y_stop, y_step = check_and_create(y_slice)
    z_start, z_stop, z_step = check_and_create(z_slice)

    i = 0
    k = 0
    async for image in producer:
        if z_start <= i and (not z_stop or i < z_stop):
            if k % z_step == 0:
                yield np.copy(image[y_start:y_stop:y_step, x_start:x_stop:x_step])
            k += 1
        i += 1


async def average_images(producer):
    """
    average_images(producer)

    Average images as they come from *producer*.
    """
    average = None
    i = 0

    async for data in producer:
        if average is None:
            average = np.zeros_like(data, dtype=np.float32)
        average = (average * i + data) / (i + 1)
        yield average
        i += 1


async def flat_correct(flat, producer, dark=None):
    """
    flat_correct(flat, producer, dark=None)

    Flat correcting coroutine which takes a *flat* field and a *dark* field (if given) from
    *producer* and calculates a flat corrected radiograph.
    """
    flat = flat.astype(np.float32)

    async for radio in producer:
        yield make_flat_correct(radio, flat, dark=dark)


async def absorptivity(producer):
    r"""
    absorptivity(producer)

    Get the absorptivity from a flat corrected stream of images.  The intensity
    after the object is defined as :math:`I = I_0 \cdot e^{-\mu t}` and we
    extract the absorptivity :math:`\mu t` from the stream of flat corrected
    images :math:`I / I_0`.
    """
    async for frame in producer:
        yield -np.log(frame)


async def stall(producer, per_shot=10, flush_at=None):
    """
    stall(producer, per_shot=10, flush_at=None)

    Send items once enough is collected from *producer*. Collect *per_shot* items. The incoming data
    might represent a collection of some kind. If the last item is supposed to be sent regardless
    the current number of collected items, use *flush_at* by which you specify the collection size
    and every time the current item *counter* % *flush_at* == 0 the item
    is sent.
    """
    i = 1

    async for item in producer:
        if i % per_shot == 0:
            yield item
        elif flush_at and i % flush_at == 0:
            yield item
            i = 0

        i += 1


class Timer(object):

    """
    Timer object measures execution times of coroutine-based workflows. It measures the time from
    when this object receives data until all the subsequent stages finish.
    """

    def __init__(self):
        self.durations = []

    def reset(self):
        """Reset the timer."""
        self.durations = []

    @property
    def duration(self):
        """All iterations summed up."""
        return sum(self.durations)

    @property
    def mean(self):
        """Mean iteration execution time."""
        return self.duration / len(self.durations)

    async def __call__(self, producer):
        """Measures the execution time of *producer*."""
        async for item in producer:
            start = time.perf_counter()
            yield item
            self.durations.append((time.perf_counter() - start) * q.s)
